"""Test PyTorchBitfountModel."""
import os
from pathlib import Path
from typing import Type
from unittest.mock import Mock

import numpy as np
import pytest
from pytest import fixture, raises
from pytest_mock import MockerFixture
import pytorch_lightning as pl
import sqlalchemy
import torch
from torchmetrics.functional import jaccard_index

from bitfount import DatabaseConnection
from bitfount.backends.pytorch.models.bitfount_model import PyTorchBitfountModel
from bitfount.data.datasets import _BitfountDataset
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datasplitters import PercentageSplitter
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import DataSplit
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.modeller import _Modeller
from bitfount.hub.api import BitfountHub
from bitfount.metrics import (
    BINARY_CLASSIFICATION_METRICS,
    MetricCollection,
    MetricsProblem,
)
from bitfount.models.base_models import ModelContext
from bitfount.utils import _get_non_abstract_classes_from_module
from tests.bitfount import TEST_SECURITY_FILES
from tests.bitfount.backends.pytorch.helper import get_params_mean
from tests.bitfount.models.test_models import SERIALIZED_MODEL_NAME
from tests.utils.helper import (
    AUC_THRESHOLD,
    TABLE_NAME,
    backend_test,
    create_dataset,
    create_datasource,
    create_datastructure,
    create_query_datastructure,
    create_schema,
    create_segmentation_dataset,
    integration_test,
    unit_test,
)


@backend_test
class TestPyTorchBitfountModel:
    """Test PyTorchBitfountModel class."""

    @fixture
    def datastructure(self) -> DataStructure:
        """Fixture for datastructure."""
        return create_datastructure()

    @fixture
    def query_datastructure(self) -> DataStructure:
        """Fixture for datastructure containing query."""
        return create_query_datastructure()

    @fixture
    def datasource(self) -> BaseSource:
        """Fixture for datasource."""
        return create_datasource(classification=True)

    @fixture
    def dummy_model_class(
        self, pytorch_bitfount_model_correct_structure: str, tmp_path: Path
    ) -> type:
        """Returns Dummy PytorchBitfountModel class."""
        model_file = tmp_path / "DummyModel.py"
        model_file.touch()
        model_file.write_text(pytorch_bitfount_model_correct_structure)
        return _get_non_abstract_classes_from_module(model_file)["DummyModel"]

    @fixture
    def dummy_model_class_tab_and_img(
        self, pytorch_bitfount_model_tab_image_data: str, tmp_path: Path
    ) -> type:
        """Returns Dummy PytorchBitfountModel class."""
        model_file = tmp_path / "DummyModelTabImg.py"
        model_file.touch()
        model_file.write_text(pytorch_bitfount_model_tab_image_data)
        return _get_non_abstract_classes_from_module(model_file)["DummyModelTabImg"]

    @integration_test
    def test_dummy_model_works_correctly(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Test fit() method runs without failure."""
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        model._pl_trainer = pl.Trainer(fast_dev_run=True)
        model.fit(datasource)

    @unit_test
    def test_dummy_model_img_tab_works_correctly(
        self,
        dummy_model_class_tab_and_img: type,
    ) -> None:
        """Test fit() method runs without failure for image&tabular tabular dataset."""
        data = create_dataset(classification=True, image=True)
        ds = DataFrameSource(data[:100])
        model = dummy_model_class_tab_and_img(
            datastructure=DataStructure(
                target="TARGET", image_cols=["image"], table=TABLE_NAME
            ),
            schema=create_schema(classification=True),
        )
        model._pl_trainer = pl.Trainer(fast_dev_run=True)
        model.fit(ds)

    @integration_test
    def test_dummy_model_learns(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mocker: MockerFixture,
    ) -> None:
        """Test that the model learns by checking metrics."""
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        model.fit(datasource)
        preds, target = model.evaluate()

        # TODO: [BIT-1604] Remove these assert statements once they become superfluous.
        assert isinstance(preds, np.ndarray)
        assert isinstance(target, np.ndarray)

        metrics = MetricCollection.create_from_model(model)
        results = metrics.compute(target, preds)
        assert isinstance(results, dict)
        assert len(metrics.metrics) == len(BINARY_CLASSIFICATION_METRICS)
        assert results["AUC"] > AUC_THRESHOLD

    @unit_test
    def test_dummy_model_evaluate(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mocker: MockerFixture,
    ) -> None:
        """Test that the evaluate sets `_test_preds` and `_test_targets`."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=create_schema(classification=True),
        )
        model.fit(datasource)
        _mock_reset_attr = mocker.patch.object(model, "_reset_test_attrs")
        model.evaluate()
        _mock_reset_attr.assert_called_once()
        assert model._test_preds is not None
        assert model._test_targets is not None

    @unit_test
    def test_dummy_model_reset_test_attrs(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Tests _reset_test_attrs sets `_test_preds` and `_test_targets` to `None`."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=create_schema(classification=True),
        )
        model._test_preds = [np.array(1), np.array(0), np.array(1), np.array(1)]
        model._test_targets = [np.array(1), np.array(1), np.array(1), np.array(0)]
        model._reset_test_attrs()
        assert model._test_preds is None
        assert model._test_targets is None

    @unit_test
    def test_init_no_classes_raises_error(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Test initialise fails with no n_classes specified and no target."""
        datasource.load_data()
        inference_datasource = DataFrameSource(datasource.data, ignore_cols=["TARGET"])
        inference_datastructure = datastructure
        inference_datastructure.target = None
        inference_model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                inference_datasource,
                ignore_cols={TABLE_NAME: ["TARGET"]},
                table_name=TABLE_NAME,
            ),
            epochs=1,
        )
        with raises(ValueError):
            inference_model.initialise_model()

    @unit_test
    def test_prediction(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Test that model prediction works after training."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
                table_name=TABLE_NAME,
            ),
            epochs=1,
        )
        model.fit(datasource)
        model.predict(datasource)

    @unit_test
    def test_initialise_with_context(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Test that model prediction works after training."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
                table_name=TABLE_NAME,
            ),
            epochs=1,
        )
        model.initialise_model(datasource, context=ModelContext.WORKER)

    @unit_test
    def test_prediction_empty_testset(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Test that model evaluation fails on empty dataset."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
                table_name=TABLE_NAME,
            ),
            epochs=1,
        )
        model.fit(datasource)
        empty_datasource = DataFrameSource(
            datasource.data, data_splitter=PercentageSplitter(0, 0)
        )
        with raises(ValueError):
            model.predict(empty_datasource)

    @integration_test
    def test_prediction_with_unsupervised_data(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mocker: MockerFixture,
        tmp_path: Path,
    ) -> None:
        """Test that model evaluation works.

        Test that predict() method works, after training and deserialization.
        Sets the n_classes field explicitly with no TARGET field present
        in the data.
        """
        model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
                table_name=TABLE_NAME,
            ),
            epochs=1,
        )
        model.fit(datasource)
        model.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        inference_datasource = DataFrameSource(datasource.data, ignore_cols=["TARGET"])
        inference_datastructure = datastructure
        inference_datastructure.target = None
        inference_datastructure.selected_cols.remove("TARGET")
        inference_datastructure._force_stype.pop("categorical")
        inference_model = dummy_model_class(
            datastructure=inference_datastructure,
            schema=BitfountSchema(
                inference_datasource,
                ignore_cols={TABLE_NAME: ["TARGET"]},
                table_name=TABLE_NAME,
            ),
            n_classes=2,
            epochs=1,
        )
        inference_model.deserialize(tmp_path / SERIALIZED_MODEL_NAME)
        preds = inference_model.predict(inference_datasource)
        mocker.patch.object(_BitfountDataset, "_set_column_name_attributes")
        mocker.patch.object(_BitfountDataset, "_reformat_data")
        dataset = _BitfountDataset(
            inference_datasource, Mock(value="test"), Mock(), Mock(), Mock()
        )

        assert preds is not None
        assert len(preds) == len(dataset.get_dataset_split(DataSplit.TEST))
        assert inference_model.n_classes == len(preds[0])

    @unit_test
    def test_serialization_before_fitting(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        tmp_path: Path,
    ) -> None:
        """Test Model can be serialized properly before fitting."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": "TARGET"}},
                table_name=TABLE_NAME,
            ),
        )
        model.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME) is True

    @unit_test
    def test_serialization_after_fitting(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        tmp_path: Path,
    ) -> None:
        """Test Model can be serialized properly after fitting."""
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        model.fit(data=datasource)
        model.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME) is True

    @unit_test
    def test_deserialization_before_fitting(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        tmp_path: Path,
    ) -> None:
        """Test Model can be deserialized properly before fitting."""
        model = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
                table_name=TABLE_NAME,
            ),
        )
        model.fit(data=datasource)
        model.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME) is True
        model2 = dummy_model_class(
            datastructure=datastructure,
            schema=BitfountSchema(
                datasource,
                force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
                table_name=TABLE_NAME,
            ),
            seed=123,
        )
        model2.deserialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert torch.isclose(
            get_params_mean(model.get_param_states()),
            get_params_mean(model2.get_param_states()),
            atol=1e-4,
        )

    @unit_test
    def test_deserialization_after_fitting(
        self,
        datasource: DataFrameSource,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        tmp_path: Path,
    ) -> None:
        """Test Model can be deserialized properly after fitting."""
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        model.fit(data=datasource)
        model.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME) is True
        model2 = dummy_model_class(
            datastructure=datastructure,
            schema=create_schema(classification=True),
            seed=123,
        )
        model2.fit(data=datasource)
        model2.deserialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert torch.isclose(
            get_params_mean(model.get_param_states()),
            get_params_mean(model2.get_param_states()),
            atol=1e-4,
        )

    @unit_test
    def test__get_import_statements(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Test _get_import_statements."""
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        assert model._get_import_statements() == [
            "import os",
            "import torch",
            "from torch import nn as nn",
            "from torch.nn import functional as F",
            "from bitfount import *",
            "import bitfount",
        ]

    @unit_test
    def test__get_model(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mocker: MockerFixture,
    ) -> None:
        """Tests private _get_model method with a custom model.

        Checks that a BitfountModelReference is created from the custom model.
        """
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        # Mock the serialization of the model
        mock_serialize_method = mocker.patch.object(
            model, "serialize_model_source_code"
        )
        extra_imports = ["from blah import blah"]
        model_ref = model._get_model(
            extra_imports=extra_imports,
            hub=Mock(spec=BitfountHub, username="test"),
        )
        assert isinstance(model_ref, BitfountModelReference)
        mock_serialize_method.assert_called_once_with(
            filename="DummyModel.py", extra_imports=extra_imports
        )

    @unit_test
    def test__fit_federated(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mock_bitfount_session: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Tests private _fit_federated method with a custom model.

        Checks that DistributedModelMixIn._fit_federated helper method creates
        correct instances and runs the modeller correctly.
        """
        # change datastructure table to be a mapping
        datastructure.table = {
            "bitfount/census-income": "census-income",
            "bitfount/census-income-2": "census-income-2",
        }
        # Create model to test
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )

        # Mock the serialization of the model
        mock_serialize_method = mocker.patch.object(
            model,
            "serialize_model_source_code",
            return_value="This is some source code",
        )

        # Patch out the modeller's run method as we only care how it is called
        # from _fit_federated.
        mock_modeller_run_method = mocker.patch.object(_Modeller, "run")

        # Run method
        pod_identifiers = ["bitfount/census-income", "bitfount/census-income-2"]
        model._fit_federated(
            pod_identifiers=pod_identifiers,
            private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
        )

        # Check run method was called correctly
        # TODO: [BIT-983] Should this check that the Modeller was instantiated
        #       correctly? Related to whether we should mock out the helper calls.
        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers, require_all_pods=False, model_out=None
        )
        mock_serialize_method.assert_called_once_with(
            filename="DummyModel.py", extra_imports=None
        )

    @unit_test
    def test__fit_federated_raises_error(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mock_bitfount_session: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Tests private _fit_federated method with a custom model.

        Checks that DistributedModelMixIn._fit_federated helper method creates
        correct instances and runs the modeller correctly.
        """
        # Create model to test
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        # Mock the serialization of the model
        mocker.patch.object(
            model,
            "serialize_model_source_code",
            return_value="This is some source code",
        )
        # Run method
        pod_identifiers = ["bitfount/census-income", "bitfount/census-income-2"]
        with pytest.raises(ValueError):
            model._fit_federated(
                pod_identifiers=pod_identifiers,
                private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
            )

    @unit_test
    def test_initialise_modeller_context(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Tests initialise_model in the modeller context."""
        datastructure.table = {"pod1": "test_table"}
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        model.initialise_model(context=ModelContext.MODELLER)

    @unit_test
    def test_model_init_with_db_query(
        self,
        db_session: sqlalchemy.engine.base.Engine,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Tests that model is initialised with a query datastructure."""
        query_datastructure = create_query_datastructure()
        model = dummy_model_class(
            datastructure=query_datastructure,
            schema=create_schema(classification=True),
        )
        db_datasource = DatabaseSource(DatabaseConnection(db_session))
        assert isinstance(query_datastructure.query, str)
        db_datasource.datastructure_query = query_datastructure.query
        expected_schema = query_datastructure._override_schema(db_datasource)
        model.initialise_model(db_datasource)
        assert model.databunch is not None
        assert model.n_classes == 2
        assert model.databunch.schema == expected_schema

    @unit_test
    def test_model_init_with_db_query_no_data(
        self,
        db_session: sqlalchemy.engine.base.Engine,
        dummy_model_class: Type[PyTorchBitfountModel],
    ) -> None:
        """Tests that model is initialised with a query datastructure."""
        query_datastructure = create_query_datastructure()
        model = dummy_model_class(
            datastructure=query_datastructure,
            schema=create_schema(classification=True),
        )
        assert isinstance(query_datastructure.query, str)
        model.initialise_model()
        assert model.input_size is not None

    @unit_test
    def test__evaluate_federated(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mock_bitfount_session: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Tests private _evaluate_federated method with a custom model.

        Checks that DistributedModelMixIn._fit_federated helper method creates
        correct instances and runs the modeller correctly.
        """
        datastructure.table = {"pod1": "pod1_table", "pod2": "pod2_table"}
        # Create model to test
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )

        # Mock the serialization of the model
        mock_serialize_method = mocker.patch.object(
            model,
            "serialize_model_source_code",
            return_value="This is some source code",
        )

        # Patch out the modeller's run method as we only care how it is called
        # from _fit_federated.
        mock_modeller_run_method = mocker.patch.object(_Modeller, "run")

        # Run method
        pod_identifiers = ["test_username/pod1", "test_username/pod2"]
        model._evaluate_federated(
            pod_identifiers=pod_identifiers,
            private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
        )

        # Check run method was called correctly
        # TODO: [BIT-983] Should this check that the Modeller was instantiated
        #       correctly? Related to whether we should mock out the helper calls.
        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers, require_all_pods=False, model_out=None
        )
        mock_serialize_method.assert_called_once_with(
            filename="DummyModel.py", extra_imports=None
        )

    @unit_test
    def test__evaluate_federated_raises_error(
        self,
        datastructure: DataStructure,
        dummy_model_class: Type[PyTorchBitfountModel],
        mock_bitfount_session: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Tests private _evaluate_federated method with a custom model.

        Checks that DistributedModelMixIn._fit_federated helper method creates
        correct instances and runs the modeller correctly.
        """
        # Create model to test
        model = dummy_model_class(
            datastructure=datastructure, schema=create_schema(classification=True)
        )

        # Mock the serialization of the model
        mocker.patch.object(
            model,
            "serialize_model_source_code",
            return_value="This is some source code",
        )

        # Patch out the modeller's run method as we only care how it is called
        # from _fit_federated.
        mocker.patch.object(_Modeller, "run")

        # Run method
        pod_identifiers = ["bitfount/census-income", "bitfount/census-income-2"]
        with pytest.raises(ValueError):
            model._evaluate_federated(
                pod_identifiers=pod_identifiers,
                private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
            )


@backend_test
class TestUNetPyTorchBitfountModel:
    """Test PyTorchBitfountModel class."""

    @fixture
    def dummy_model_class(
        self, pytorch_bitfount_segmentation_model: str, tmp_path: Path
    ) -> type:
        """Returns Dummy PytorchBitfountModel class."""
        model_file = tmp_path / "DummyUnet.py"
        model_file.touch()
        model_file.write_text(pytorch_bitfount_segmentation_model)
        return _get_non_abstract_classes_from_module(model_file)["DummyUnet"]

    @unit_test
    def test_dummy_model_works(
        self, dummy_model_class: Type[PyTorchBitfountModel], tmp_path: Path
    ) -> None:
        """Tests that the segmentation model works."""
        seg_dir = tmp_path / "seg/"
        seg_dir.mkdir()
        seg_dataset = create_segmentation_dataset(seg_dir=seg_dir)
        ds = DataStructure(
            table=TABLE_NAME, image_cols=["img", "masks"], target="masks"
        )
        datasource = DataFrameSource(seg_dataset)
        schema = BitfountSchema(
            datasource,
            force_stypes={TABLE_NAME: {"image": ["img", "masks"]}},
            table_name=TABLE_NAME,
        )
        model = dummy_model_class(
            n_channels=3,
            n_classes=3,
            datastructure=ds,
            schema=schema,
            steps=1,
            batch_size=5,
        )
        model.fit(datasource)

    @integration_test
    def test_dummy_model_learns(
        self,
        dummy_model_class: Type[PyTorchBitfountModel],
        mocker: MockerFixture,
        tmp_path: Path,
    ) -> None:
        """Tests the segmentation model learns by checking metrics."""
        seg_dir = tmp_path / "seg/"
        seg_dir.mkdir()
        seg_dataset = create_segmentation_dataset(seg_dir=seg_dir)
        ds = DataStructure(
            table=TABLE_NAME, image_cols=["img", "masks"], target="masks"
        )
        datasource = DataFrameSource(seg_dataset)
        schema = BitfountSchema(
            datasource,
            force_stypes={TABLE_NAME: {"image": ["img", "masks"]}},
            table_name=TABLE_NAME,
        )
        model = dummy_model_class(
            n_channels=3,
            n_classes=3,
            datastructure=ds,
            schema=schema,
            epochs=1,
            batch_size=5,
        )
        model.fit(datasource)
        preds, target = model.evaluate()

        # TODO: [BIT-1604] Remove these assert statements once they become superfluous.
        assert isinstance(preds, np.ndarray)
        assert isinstance(target, np.ndarray)
        metrics = MetricCollection(problem=MetricsProblem.SEGMENTATION)
        results = metrics.compute(target, preds)

        assert len(results) == 3
        # Confirm that our IoU metric matches the pytorch implementation.
        assert (
            round(
                jaccard_index(torch.tensor(preds), torch.tensor(target), num_classes=3)
                .numpy()
                .tolist(),
                4,
            )
            == results["IoU"]
        )



"""Fixtures for mocking of encryption-related functions."""
from unittest.mock import Mock

from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.transport.message_service import _MessageEncryption

_sign_sentinel: bytes = b"i_am_not_a_real_signature"


@fixture
def mock_rsa_encryption(mocker: MockerFixture) -> Mock:
    """Mock out RSAEncryption.encrypt().

    Simply returns the message passed in.

    Should be used in conjunction with the mock_rsa_decryption fixture if both
    encryption and decryption are needed.
    """
    mock_rsa_encryption: Mock = mocker.patch.object(
        _RSAEncryption,
        "encrypt",
        autospec=True,
        side_effect=lambda message, _key: message,
    )
    return mock_rsa_encryption


@fixture
def mock_rsa_decryption(mocker: MockerFixture) -> Mock:
    """Mock out RSAEncryption.decrypt().

    Simply returns the ciphertext passed in.

    Should be used in conjunction with the mock_rsa_encryption fixture if both
    encryption and decryption are needed.
    """
    mock_rsa_decryption: Mock = mocker.patch.object(
        _RSAEncryption,
        "decrypt",
        autospec=True,
        side_effect=lambda ciphertext, _key: ciphertext,
    )
    return mock_rsa_decryption


@fixture
def mock_rsa_sign_message(mocker: MockerFixture) -> Mock:
    """Mock out RSAEncryption.sign_message().

    Simply returns a sentinel signature.

    Should be used in conjunction with the mock_rsa_signature_verification fixture
    if both signing and validating are needed.
    """
    mock_rsa_sign_message: Mock = mocker.patch.object(
        _RSAEncryption,
        "sign_message",
        autospec=True,
        side_effect=lambda _key, _message: _sign_sentinel,
    )
    return mock_rsa_sign_message


@fixture
def mock_rsa_signature_verification(mocker: MockerFixture) -> Mock:
    """Mock out RSAEncryption.verify_signature().

    Returns True if the signature is the sentinel signature, False otherwise.

    Should be used in conjunction with the mock_rsa_sign_message fixture if both
    signing and validating are needed.
    """
    mock_rsa_signature_verification: Mock = mocker.patch.object(
        _RSAEncryption,
        "verify_signature",
        autospec=True,
        side_effect=lambda _key, signature, _message: bool(signature == _sign_sentinel),
    )
    return mock_rsa_signature_verification


@fixture
def mock_message_aes_encryption(mocker: MockerFixture) -> Mock:
    """Mock out MessageEncryption.encrypt_outgoing_message().

    Simply returns the message body passed in.

    Should be used in conjunction with the mock_message_aes_decryption fixture if both
    encryption and decryption are needed.
    """
    mock_message_aes_encryption: Mock = mocker.patch.object(
        _MessageEncryption,
        "encrypt_outgoing_message",
        autospec=True,
        side_effect=lambda body, _key: body,
    )
    return mock_message_aes_encryption


@fixture
def mock_message_aes_decryption(mocker: MockerFixture) -> Mock:
    """Mock out MessageEncryption.decrypt_incoming_message().

    Simply returns the message body passed in.

    Should be used in conjunction with the mock_message_aes_encryption fixture if both
    encryption and decryption are needed.
    """
    mock_message_aes_decryption: Mock = mocker.patch.object(
        _MessageEncryption,
        "decrypt_incoming_message",
        autospec=True,
        side_effect=lambda body, _key: body,
    )
    return mock_message_aes_decryption



"""Test worker can communicate with modeller."""
from __future__ import annotations

import asyncio
from collections import namedtuple
from datetime import datetime, timezone
import logging
import re
from typing import Callable, Dict, Iterable, List, Optional
from unittest.mock import ANY, Mock, NonCallableMagicMock, call, create_autospec

from _pytest.logging import LogCaptureFixture
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from grpc import RpcError
import msgpack
import pytest
from pytest import fixture, raises
from pytest_mock import MockerFixture

from bitfount.federated.transport.base_transport import (
    MessageRetrievalError,
    _run_func_and_listen_to_mailbox,
)
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _MessageEncryption,
    _MessageService,
)
from bitfount.federated.transport.protos.messages_pb2 import SuccessResponse
from bitfount.federated.transport.worker_transport import (
    _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT,
    _SOFT_LIMIT_MESSAGE_TIMEOUT,
    _get_psi_dataset,
    _get_worker_secure_shares,
    _InterPodWorkerMailbox,
    _send_secure_shares_to_others,
    _WorkerMailbox,
)
from bitfount.federated.types import _PodResponseType
from tests.utils import PytestRequest
from tests.utils.fixtures import PollForMessagesPatcher
from tests.utils.helper import (
    get_debug_logs,
    get_error_logs,
    get_info_logs,
    get_warning_logs,
    unit_test,
)

WorkerDetails = namedtuple("WorkerDetails", ["pod_identifier", "mailbox_id"])


@fixture
def pod_identifier() -> str:
    """A pod identifier."""
    return "someUser/somePod"


@fixture
def worker_mailbox_id() -> str:
    """A mailbox ID for the worker."""
    return "thisWorkersMailboxID"


@fixture
def modeller_mailbox_id() -> str:
    """The mailbox ID for the modeller."""
    return "someModellerMailboxId"


@fixture
def modeller_name() -> str:
    """The name of the modeller."""
    return "someModeller"


@fixture
def aes_key() -> bytes:
    """The AES key to use in message encryption between worker and modeller."""
    return b"SecretAESKey"


@fixture
def mock_message_service() -> Mock:
    """Mock message service."""
    mock_message_service: Mock = create_autospec(_MessageService, instance=True)
    return mock_message_service


@fixture
def other_pod_details() -> List[WorkerDetails]:
    """Details of the other pods involved in the task."""
    return [
        WorkerDetails("user/differentpod", "someMailboxID2"),
        WorkerDetails("another/pod", "someMailboxID3"),
    ]


@fixture
def pod_mailbox_ids(
    other_pod_details: List[WorkerDetails], pod_identifier: str, worker_mailbox_id: str
) -> Dict[str, str]:
    """A mapping of pod identifier to mailbox ID for all task workers.

    This includes the worker itself.
    """
    return {
        pod_identifier: worker_mailbox_id,
        other_pod_details[0].pod_identifier: other_pod_details[0].mailbox_id,
        other_pod_details[1].pod_identifier: other_pod_details[1].mailbox_id,
    }


@fixture(params=(True, False), ids=("task_id_incl", "no_task_id"))
def opt_task_id(request: PytestRequest) -> Optional[str]:
    """Returns a task ID or None, to cover both cases."""
    incl_task_id: bool = request.param
    if incl_task_id:
        return "this-is-a-task-id"
    else:
        return None


@fixture
def online_check_uuid() -> str:
    """UUID for online checking."""
    return "auuidforonlinecheck"


@fixture
def worker_mailbox(
    aes_key: bytes,
    mock_message_service: Mock,
    modeller_mailbox_id: str,
    modeller_name: str,
    opt_task_id: Optional[str],
    pod_identifier: str,
    pod_mailbox_ids: Dict[str, str],
) -> _WorkerMailbox:
    """A WorkerMailbox instance with components mocked out."""
    return _WorkerMailbox(
        pod_identifier=pod_identifier,
        modeller_mailbox_id=modeller_mailbox_id,
        modeller_name=modeller_name,
        aes_encryption_key=aes_key,
        message_service=mock_message_service,
        pod_mailbox_ids=pod_mailbox_ids,
        task_id=opt_task_id,
    )


@fixture
def mock_pod_public_keys(other_pod_details: List[WorkerDetails]) -> Dict[str, Mock]:
    """Mock public keys for each "other" pod."""
    other_pod_ids = [wd.pod_identifier for wd in other_pod_details]
    return {
        pod_id: create_autospec(RSAPublicKey, instance=True) for pod_id in other_pod_ids
    }


@fixture
def mock_private_key() -> Mock:
    """Mock private key for target pod."""
    mock_private_key: Mock = create_autospec(RSAPrivateKey, instance=True)
    return mock_private_key


@fixture
def interpod_worker_mailbox(
    aes_key: bytes,
    mock_message_service: Mock,
    mock_pod_public_keys: Dict[str, Mock],
    mock_private_key: Mock,
    modeller_mailbox_id: str,
    modeller_name: str,
    opt_task_id: Optional[str],
    other_pod_details: List[WorkerDetails],
    pod_identifier: str,
    pod_mailbox_ids: Dict[str, str],
) -> _InterPodWorkerMailbox:
    """An InterPodWorkerMailbox instance with components mocked out."""
    return _InterPodWorkerMailbox(
        pod_public_keys=mock_pod_public_keys,
        private_key=mock_private_key,
        pod_identifier=pod_identifier,
        modeller_mailbox_id=modeller_mailbox_id,
        modeller_name=modeller_name,
        aes_encryption_key=aes_key,
        message_service=mock_message_service,
        pod_mailbox_ids=pod_mailbox_ids,
        task_id=opt_task_id,
    )


@fixture
def mock_message_type() -> Mock:
    """A mock instance of the _BitfountMessageType enum."""
    mock_message_type: Mock = create_autospec(_BitfountMessageType, instance=True)
    return mock_message_type


@fixture
def dummy_message_awaitable() -> Callable[
    [Optional[_BitfountMessage], Optional[bool]], object
]:
    """A dummy awaitable for testing."""

    class MockMessageAwaitable:
        """Dummy awaitable in place of _AsyncCallback.

        The `timeout` indicates whether the first time the awaitable is called,
        it should return a message or raise an `asyncio.TimeoutError`.
        """

        def __init__(
            self,
            message: Optional[_BitfountMessage] = None,
            timeout: Optional[bool] = False,
        ) -> None:
            self.message = message
            self.timeout = timeout
            self.timeout_handler = Mock()

        async def result(
            self, timeout: Optional[float] = None
        ) -> Optional[_BitfountMessage]:
            """Returns the result of the awaitable.

            In this dummy implementation, the result is just the original message. If a
            timeout is specified, this will be recorded using `self.timeout_handler`.
            """
            if self.timeout:
                self.timeout = None
                raise asyncio.TimeoutError()

            self.timeout_handler(timeout)
            return self.message

    return MockMessageAwaitable


@unit_test
class TestWorkerMailbox:
    """Test WorkerMailbox class."""

    async def test__send_aes_encrypted_message_successful(
        self,
        aes_key: bytes,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        pod_identifier: str,
        worker_mailbox: _WorkerMailbox,
        worker_mailbox_id: str,
    ) -> None:
        """Message is sent AES encrypted using provided key."""
        fake_timestamps = ["Hello"]
        mock_message_timestamps(fake_timestamps)

        expected_encrypted_message: bytes = b"encrypted_message"
        encrypt = mocker.patch.object(_MessageEncryption, "encrypt_outgoing_message")
        encrypt.return_value = expected_encrypted_message

        message = "some message"
        dumped_message = msgpack.dumps(message)

        await worker_mailbox._send_aes_encrypted_message(
            message, _BitfountMessageType.TRAINING_UPDATE
        )

        encrypt.assert_called_once_with(dumped_message, aes_key)
        mock_message_service.send_message.assert_called_once_with(
            _BitfountMessage(
                message_type=_BitfountMessageType.TRAINING_UPDATE,
                body=expected_encrypted_message,
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=pod_identifier,
                sender_mailbox_id=worker_mailbox_id,
                timestamp=fake_timestamps[0],
                task_id=opt_task_id,
            ),
            already_packed=True,
        )

    async def test__send_aes_encrypted_message_throws_error(
        self,
        aes_key: bytes,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        pod_identifier: str,
        rpc_error: RpcError,
        worker_mailbox: _WorkerMailbox,
        worker_mailbox_id: str,
    ) -> None:
        """Sending encrypted message receives RpcError."""
        fake_timestamps = ["Hello"]
        mock_message_timestamps(fake_timestamps)

        expected_encrypted_message: bytes = b"encrypted_message"
        encrypt = mocker.patch.object(_MessageEncryption, "encrypt_outgoing_message")
        encrypt.return_value = expected_encrypted_message

        message = "some message"
        dumped_message = msgpack.dumps(message)

        mock_message_service.send_message.side_effect = rpc_error

        with raises(RpcError):
            await worker_mailbox._send_aes_encrypted_message(
                message, _BitfountMessageType.TRAINING_UPDATE
            )

        encrypt.assert_called_once_with(dumped_message, aes_key)
        mock_message_service.send_message.assert_called_once_with(
            _BitfountMessage(
                message_type=_BitfountMessageType.TRAINING_UPDATE,
                body=expected_encrypted_message,
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=pod_identifier,
                sender_mailbox_id=worker_mailbox_id,
                timestamp=fake_timestamps[0],
                task_id=opt_task_id,
            ),
            already_packed=True,
        )

    async def test__get_message_returns_message_without_breaching_timeout(
        self,
        caplog: LogCaptureFixture,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        pod_identifier: str,
        worker_mailbox: _WorkerMailbox,
        worker_mailbox_id: str,
    ) -> None:
        """Message is returned as expected without breaching timeout."""
        # Set up the mocked messages to be received
        patch_poll_for_messages(
            [
                _BitfountMessage(
                    message_type=_BitfountMessageType.TRAINING_UPDATE,
                    body=b"123",
                    recipient=modeller_name,
                    recipient_mailbox_id=modeller_mailbox_id,
                    sender=pod_identifier,
                    sender_mailbox_id=worker_mailbox_id,
                    task_id=opt_task_id,
                )
            ]
        )

        # Start listening and processing messages
        mock_soft_limit_timeout = 30
        message = await _run_func_and_listen_to_mailbox(
            worker_mailbox._get_message(
                _BitfountMessageType.TRAINING_UPDATE, timeout=mock_soft_limit_timeout
            ),
            worker_mailbox,
        )
        assert message.body == b"123"

        # Check no logs due to timeout not being reached
        logs = get_info_logs(caplog, and_higher=True)
        assert "Checking if Modeller is still online..." not in logs

    async def test__get_message_breaches_soft_limit_timeout(
        self,
        caplog: LogCaptureFixture,
        dummy_message_awaitable: Callable[
            [Optional[_BitfountMessage], Optional[bool]], object
        ],
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        modeller_name: str,
        online_check_uuid: str,
        opt_task_id: Optional[str],
        pod_identifier: str,
        worker_mailbox: _WorkerMailbox,
        worker_mailbox_id: str,
    ) -> None:
        """Message is returned as expected after breaching soft timeout."""
        caplog.set_level("INFO")
        original_message_awaitable = dummy_message_awaitable(
            _BitfountMessage(
                message_type=_BitfountMessageType.TRAINING_UPDATE,
                body=b"123",
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=pod_identifier,
                sender_mailbox_id=worker_mailbox_id,
                task_id=opt_task_id,
            ),
            True,
        )

        # Patch out the waiting for response
        mocker.patch.object(
            worker_mailbox._online_response_handler,
            "wait_for_response",
            autospec=True,
            return_value=None,
        )

        # Mock the bitfount messages retrieved from the modeller
        mocker.patch(
            "bitfount.federated.transport.worker_transport._get_message_awaitable",
            return_value=original_message_awaitable,
        )

        # Mock checking the modeller is online
        mock_check_modeller_online = mocker.patch.object(
            worker_mailbox, "check_modeller_online"
        )

        # Mock hard limit timeout
        mocker.patch(
            "bitfount.federated.transport.worker_transport._HARD_LIMIT_MESSAGE_TIMEOUT",
            5,
        )

        # Start listening and processing messages
        mock_soft_limit_timeout = 3
        message = await worker_mailbox._get_message(
            _BitfountMessageType.TRAINING_UPDATE, timeout=mock_soft_limit_timeout
        )

        assert message.body == b"123"

        # Check logs
        info_logs = get_info_logs(caplog)
        assert "Checking if Modeller is still online..." in info_logs
        assert "Modeller is online, responded with expected message." in info_logs

        mock_check_modeller_online.assert_called_once()

        # Should be one call to timeout_handler; the first call to result() raises
        # a TimeoutError, the second call passes the timeout through to
        # timeout_handler. This second call is made whilst we are waiting for the
        # online response check event to be set and so has no timeout.
        # mypy reason: Mypy can't detect the actual type of this mocked awaitable.
        original_message_awaitable.timeout_handler.assert_called_once()  # type: ignore[attr-defined] # Reason: see above # noqa: B950

    async def test__get_message_breaches_hard_limit_timeout(
        self,
        caplog: LogCaptureFixture,
        dummy_message_awaitable: Callable[
            [Optional[_BitfountMessage], Optional[bool]], object
        ],
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Test worker gives up on waiting for message after breaching hard limit.

        This should raise an `asyncio.TimeoutError` exception.
        """
        caplog.set_level("INFO")

        # No messages retrieved from the Modeller, just `TimeOutError`s
        original_message_awaitable = dummy_message_awaitable(None, True)

        # Mock the bitfount messages retrieved from the modeller
        mocker.patch(
            "bitfount.federated.transport.worker_transport._get_message_awaitable",
            return_value=original_message_awaitable,
        )

        # Mock checking the modeller is online
        mock_check_modeller_online = mocker.patch.object(
            worker_mailbox, "check_modeller_online"
        )

        # Mock sending of task abort message
        mock_send_task_abort_message = mocker.patch.object(
            worker_mailbox, "send_task_abort_message"
        )

        # Mock hard limit timeout
        mocker.patch(
            "bitfount.federated.transport.worker_transport._HARD_LIMIT_MESSAGE_TIMEOUT",
            5,
        )
        mock_wait_for = mocker.patch.object(
            asyncio, "wait_for", autospec=True, side_effect=asyncio.TimeoutError()
        )

        # Start listening and processing messages
        mock_soft_limit_timeout = 3
        with pytest.raises(asyncio.TimeoutError):
            await worker_mailbox._get_message(
                _BitfountMessageType.TRAINING_UPDATE, timeout=mock_soft_limit_timeout
            )

        # Check logs
        assert "Checking if Modeller is still online..." in get_info_logs(caplog)
        assert "Modeller is offline. Aborting task." in get_warning_logs(caplog)

        mock_check_modeller_online.assert_called_once()
        mock_wait_for.assert_awaited_once_with(ANY, timeout=5)
        mock_send_task_abort_message.assert_called_once()

    async def test__get_message_and_decrypt_successful(
        self,
        aes_key: bytes,
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        pod_identifier: str,
        worker_mailbox: _WorkerMailbox,
        worker_mailbox_id: str,
    ) -> None:
        """Receives and decrypts messages using provided AES key."""
        expected_message = "some message"
        dumped_message: bytes = msgpack.dumps(expected_message)
        # We will be manually mocking out the decryption so this message can just
        # be arbitrary.
        mock_encrypted_message: bytes = b"encrypted_message"

        # Mock out decryption
        decrypt = mocker.patch.object(_MessageEncryption, "decrypt_incoming_message")
        decrypt.return_value = dumped_message

        # Set up the mocked messages to be received
        patch_poll_for_messages(
            [
                _BitfountMessage(
                    message_type=_BitfountMessageType.TRAINING_UPDATE,
                    body=mock_encrypted_message,
                    recipient=modeller_name,
                    recipient_mailbox_id=modeller_mailbox_id,
                    sender=pod_identifier,
                    sender_mailbox_id=worker_mailbox_id,
                    task_id=opt_task_id,
                )
            ]
        )

        # Start listening and processing messages
        message = await _run_func_and_listen_to_mailbox(
            worker_mailbox._get_message_and_decrypt(
                _BitfountMessageType.TRAINING_UPDATE
            ),
            worker_mailbox,
        )

        assert message == expected_message
        decrypt.assert_called_once_with(mock_encrypted_message, aes_key)

    async def test__get_message_and_decrypt_error(
        self,
        patch_poll_for_messages: PollForMessagesPatcher,
        rpc_error: RpcError,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Receiving message throws RpcError."""
        patch_poll_for_messages([rpc_error])

        with raises(
            MessageRetrievalError,
            match="An error occurred when trying to communicate with the "
            "messaging service",
        ):
            # Start listening and processing messages
            await _run_func_and_listen_to_mailbox(
                worker_mailbox._get_message_and_decrypt(
                    _BitfountMessageType.TRAINING_UPDATE  # this is arbitrary
                ),
                worker_mailbox,
            )

    async def test_accept_job_successful(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Training job acceptance sent."""
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            return_value=SuccessResponse(),
        )

        await worker_mailbox.accept_task()

        mock_message_send.assert_called_once_with(
            {_PodResponseType.ACCEPT.name: worker_mailbox.pod_identifier},
            _BitfountMessageType.JOB_ACCEPT,
        )

    async def test_accept_job_unsuccessful(
        self,
        mocker: MockerFixture,
        rpc_error: RpcError,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Sending training job acceptance fails."""
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            side_effect=rpc_error,
        )

        with pytest.raises(RpcError):
            await worker_mailbox.accept_task()

        mock_message_send.assert_called_once_with(
            {_PodResponseType.ACCEPT.name: worker_mailbox.pod_identifier},
            _BitfountMessageType.JOB_ACCEPT,
        )

    async def test_reject_job_successful(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Training job rejection sent."""
        expected_error_messages = {"error": "messages"}
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            return_value=SuccessResponse(),
        )

        await worker_mailbox.reject_task(
            expected_error_messages,
        )

        mock_message_send.assert_called_once_with(
            expected_error_messages,
            _BitfountMessageType.JOB_REJECT,
        )

    async def test_reject_job_unsuccessful(
        self,
        mocker: MockerFixture,
        rpc_error: RpcError,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Sending training job rejection fails."""
        expected_error_messages = {"error": "messages"}
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            side_effect=rpc_error,
        )

        with pytest.raises(RpcError):
            await worker_mailbox.reject_task(
                expected_error_messages,
            )

        mock_message_send.assert_called_once_with(
            expected_error_messages,
            _BitfountMessageType.JOB_REJECT,
        )

    async def test_issue_saml_challenge_successful(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Training SAML challenge sent."""
        expected_saml_request = "some saml request"
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            return_value=SuccessResponse(),
        )

        await worker_mailbox.issue_saml_challenge(
            expected_saml_request,
        )

        mock_message_send.assert_called_once_with(
            expected_saml_request,
            _BitfountMessageType.SAML_REQUEST,
        )

    async def test_issue_saml_challenge_unsuccessful(
        self,
        mocker: MockerFixture,
        rpc_error: RpcError,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Sending SAML challenge fails."""
        expected_saml_request = "some saml request"
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            side_effect=rpc_error,
        )

        with pytest.raises(RpcError):
            await worker_mailbox.issue_saml_challenge(
                expected_saml_request,
            )

        mock_message_send.assert_called_once_with(
            expected_saml_request,
            _BitfountMessageType.SAML_REQUEST,
        )

    async def test_get_saml_response(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """SAML challenge response retrieved."""
        expected_saml_response = "some saml response"
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_saml_response,
        )

        response = await worker_mailbox.get_saml_response()

        assert response == expected_saml_response
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.SAML_RESPONSE,
            _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT,
        )

    async def test_get_task_start_update(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests worker gets task start empty message."""
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=None,
        )

        await worker_mailbox.get_task_start_update()
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.TASK_START, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
        )

    async def test_get_task_complete_update(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests worker gets task complete empty message."""
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=None,
        )

        await worker_mailbox.get_task_complete_update()
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.TASK_COMPLETE, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
        )

    async def test_get_training_complete_update(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests worker gets training complete empty message."""
        expected_message = True
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        response = await worker_mailbox.get_training_iteration_complete_update()
        assert response == expected_message
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.TRAINING_COMPLETE, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
        )

    async def test_send_oidc_client_id(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Test send_oidc_client_id() works correctly."""
        # Patch out _WorkerMailbox._send_aes_encrypted_message()
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            return_value=SuccessResponse(),
        )

        client_id = "client_id_value"

        await worker_mailbox.send_oidc_client_id(client_id)

        # Check called correctly
        mock_message_send.assert_called_once_with(
            {"client_id": client_id},
            _BitfountMessageType.OIDC_CHALLENGE,
        )

    async def test_get_oidc_auth_flow_response(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests get_oidc_auth_flow_response() correctly extracts details."""
        expected_message = {
            "auth_code": "auth_code_value",
            "code_verifier": "code_verifier_value",
            "redirect_uri": "redirect_uri_value",
        }

        # Patch out _WorkerMailbox._get_message_and_decrypt()
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        response = await worker_mailbox.get_oidc_auth_flow_response()

        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.OIDC_AFC_PKCE_RESPONSE, ANY
        )
        assert response.auth_code == expected_message["auth_code"]
        assert response.code_verifier == expected_message["code_verifier"]
        assert response.redirect_uri == expected_message["redirect_uri"]

    async def test_get_oidc_auth_flow_response_fails_wrong_type(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests get_oidc_auth_flow_response() fails with wrong type."""
        expected_message = "not a dict"

        # Patch out _WorkerMailbox._get_message_and_decrypt()
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"Unable to access OIDC response contents; "
                f"expected dict, got {type(expected_message)}"
            ),
        ):
            await worker_mailbox.get_oidc_auth_flow_response()

        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.OIDC_AFC_PKCE_RESPONSE, ANY
        )

    @pytest.mark.parametrize(
        "key_to_drop", ("auth_code", "code_verifier", "redirect_uri")
    )
    async def test_get_oidc_auth_flow_response_fails_missing_key(
        self,
        key_to_drop: str,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests get_oidc_auth_flow_response() fails with missing key."""
        expected_message = {
            "auth_code": "auth_code_value",
            "code_verifier": "code_verifier_value",
            "redirect_uri": "redirect_uri_value",
        }
        expected_message.pop(key_to_drop)

        # Patch out _WorkerMailbox._get_message_and_decrypt()
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        with pytest.raises(
            KeyError,
            match=re.escape(
                f"Expected auth_code, code_verifier, and redirect_uri to be in "
                f"OIDC response; got {expected_message.keys()}"
            ),
        ):
            await worker_mailbox.get_oidc_auth_flow_response()

        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.OIDC_AFC_PKCE_RESPONSE, ANY
        )

    async def test_get_oidc_device_code_response(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests get_oidc_device_code_response() correctly extracts details."""
        now = datetime.now(timezone.utc)
        expected_message = {
            "device_code": "someDeviceCode",
            "expires_at": now.isoformat(),
            "interval": 5,
        }

        # Patch out _WorkerMailbox._get_message_and_decrypt()
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        response = await worker_mailbox.get_oidc_device_code_response()

        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE, ANY
        )
        assert response.device_code == expected_message["device_code"]
        assert response.expires_at == now
        assert response.interval == 5

    async def test_get_oidc_device_code_response_fails_wrong_type(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests get_oidc_device_code_response() fails with wrong type."""
        expected_message = "not a dict"

        # Patch out _WorkerMailbox._get_message_and_decrypt()
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"Unable to access OIDC response contents; "
                f"expected dict, got {type(expected_message)}"
            ),
        ):
            await worker_mailbox.get_oidc_device_code_response()

        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE, ANY
        )

    @pytest.mark.parametrize(
        "key_to_drop",
        ("device_code", "expires_at", "interval"),
    )
    async def test_get_oidc_device_code_response_fails_missing_key(
        self,
        key_to_drop: str,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests get_oidc_device_code_response() fails with missing key."""
        now = datetime.now(timezone.utc)
        expected_message = {
            "device_code": "someDeviceCode",
            "expires_at": now.isoformat(),
            "interval": 5,
        }
        expected_message.pop(key_to_drop)

        # Patch out _WorkerMailbox._get_message_and_decrypt()
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )

        with pytest.raises(
            KeyError,
            match=re.escape(
                f"Expected device_code, expires_at, and interval to be in "
                f"OIDC response; got {expected_message.keys()}"
            ),
        ):
            await worker_mailbox.get_oidc_device_code_response()

        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE, ANY
        )

    async def test__get_psi_dataset(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Test _get_psi_dataset method."""
        # Patch out _WorkerMailbox._get_message_and_decrypt()
        expected_message = ["some message"]
        mock_message_decrypt = mocker.patch.object(
            worker_mailbox,
            "_get_message_and_decrypt",
            return_value=expected_message,
        )
        await _get_psi_dataset(worker_mailbox)
        # Check calls/returns
        mock_message_decrypt.assert_awaited_once_with(
            _BitfountMessageType.PSI_DATASET, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
        )

    async def test_log(
        self, mocker: MockerFixture, worker_mailbox: _WorkerMailbox
    ) -> None:
        """Test that log message is sent appropriately."""
        mock_log = mocker.patch.object(worker_mailbox, "_send_aes_encrypted_message")
        await worker_mailbox.log({"msg": "message"})
        mock_log.assert_awaited_once_with(
            {"msg": "message"}, _BitfountMessageType.LOG_MESSAGE
        )


@unit_test
class TestInterPodWorkerMailbox:
    """Tests for _InterPodWorkerMailbox."""

    def test_init_fails_if_missing_keys(
        self,
        aes_key: bytes,
        mock_message_service: Mock,
        mock_pod_public_keys: Dict[str, Mock],
        mock_private_key: Mock,
        modeller_mailbox_id: str,
        modeller_name: str,
        other_pod_details: List[WorkerDetails],
        pod_identifier: str,
        pod_mailbox_ids: Dict[str, str],
    ) -> None:
        """Tests that init raises exception if missing public keys for some pods."""
        # Remove public key from dictionary
        missing_key_pod_id = other_pod_details[-1].pod_identifier
        del mock_pod_public_keys[missing_key_pod_id]

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"We are missing public keys for the following pods: "
                f"{missing_key_pod_id}. "
                f"Unable to continue inter-pod communication."
            ),
        ):
            _InterPodWorkerMailbox(
                pod_public_keys=mock_pod_public_keys,
                private_key=mock_private_key,
                pod_identifier=pod_identifier,
                modeller_mailbox_id=modeller_mailbox_id,
                modeller_name=modeller_name,
                aes_encryption_key=aes_key,
                message_service=mock_message_service,
                pod_mailbox_ids=pod_mailbox_ids,
            )

    async def test__send_pod_to_pod_message(
        self,
        interpod_worker_mailbox: _InterPodWorkerMailbox,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mock_message_type: Mock,
        mock_pod_public_keys: Dict[str, Mock],
        mock_rsa_encryption: Mock,
        opt_task_id: Optional[str],
        other_pod_details: List[WorkerDetails],
        pod_identifier: str,
        worker_mailbox_id: str,
    ) -> None:
        """Test _send_pod_to_pod_message method works."""
        fake_timestamps = ["Hello"]
        mock_message_timestamps(fake_timestamps)

        msg_body = {"a": "message", "to": "send"}
        recipient_details = other_pod_details[0]
        recipient_key = mock_pod_public_keys[recipient_details.pod_identifier]

        await interpod_worker_mailbox._send_pod_to_pod_message(
            recipient=recipient_details.pod_identifier,
            recipient_mailbox_id=recipient_details.mailbox_id,
            object_to_send=msg_body,
            message_type=mock_message_type,
        )

        # Check encryption called correctly
        mock_rsa_encryption.assert_called_once_with(
            msgpack.dumps(msg_body), recipient_key
        )
        # Check expected message sent
        expected_message = _BitfountMessage(
            message_type=mock_message_type,
            body=msgpack.dumps(msg_body),  # mock encryption just returns input
            recipient=recipient_details.pod_identifier,
            recipient_mailbox_id=recipient_details.mailbox_id,
            sender=pod_identifier,
            sender_mailbox_id=worker_mailbox_id,
            timestamp=fake_timestamps[0],
            task_id=opt_task_id,
        )
        mock_message_service.send_message.assert_awaited_once_with(
            expected_message, already_packed=True
        )

    async def test__send_pod_to_pod_message_logs_error_if_no_key(
        self,
        caplog: LogCaptureFixture,
        interpod_worker_mailbox: _InterPodWorkerMailbox,
        mock_message_service: Mock,
        mock_message_type: Mock,
    ) -> None:
        """Test _send_pod_to_pod_message logs error if no public key for recipient."""
        await interpod_worker_mailbox._send_pod_to_pod_message(
            recipient="not_a_real_recipient/pod",
            recipient_mailbox_id="not_a_real_mailbox_id",
            object_to_send=Mock(),
            message_type=mock_message_type,
        )

        # Check error logged
        error_logs = get_error_logs(caplog)
        assert (
            "Unable to find public key for pod not_a_real_recipient/pod. "
            "Unable to send pod-to-pod message." in error_logs
        )
        # Check didn't try to send
        mock_message_service.send_message.assert_not_called()

    def test__pod_to_pod_message_handler(
        self,
        interpod_worker_mailbox: _InterPodWorkerMailbox,
        mock_private_key: Mock,
    ) -> None:
        """Test _pod_to_pod_message_handler returns decrypted RSA message."""
        mock_message = create_autospec(_BitfountMessage, instance=True)

        decrypted = interpod_worker_mailbox._pod_to_pod_message_handler(mock_message)

        # Check message decrypted correctly
        mock_message.decrypt_rsa.assert_called_once_with(mock_private_key)
        assert decrypted == mock_message.decrypt_rsa.return_value

    async def test_check_modeller_online(
        self,
        mocker: MockerFixture,
        online_check_uuid: str,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests worker `check_modeller_online` method."""
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            return_value=None,
        )

        await worker_mailbox.check_modeller_online(online_check_uuid)
        mock_message_send.assert_awaited_once_with(
            online_check_uuid,
            _BitfountMessageType.ONLINE_CHECK,
        )

    async def test_send_task_abort_message(
        self,
        mocker: MockerFixture,
        worker_mailbox: _WorkerMailbox,
    ) -> None:
        """Tests worker `send_task_abort_message` method."""
        mock_message_send = mocker.patch.object(
            worker_mailbox,
            "_send_aes_encrypted_message",
            return_value=None,
        )

        await worker_mailbox.send_task_abort_message()
        mock_message_send.assert_awaited_once_with(
            None, _BitfountMessageType.TASK_ABORT
        )


@unit_test
class TestWorkerTransportFunctions:
    """Test the other functions not contained within WorkerMailbox."""

    async def test_send_to_pods_sends_message(
        self,
        interpod_worker_mailbox: _InterPodWorkerMailbox,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mock_rsa_encryption: Mock,
        opt_task_id: Optional[str],
        other_pod_details: List[WorkerDetails],
        pod_identifier: str,
        worker_mailbox_id: str,
    ) -> None:
        """Tests send_secure_shares_to_others method works correctly.

        Asserts each pod receives their own message generated by secure_share_generator.
        """
        fake_timestamps = ["Hello", "World"]
        mock_message_timestamps(fake_timestamps)

        mock_message_service.send_message.return_value = SuccessResponse()

        class DummySecureShare:
            def __init__(self) -> None:
                self.counter = 0

            def message_body_generator(self) -> int:
                self.counter += 1
                return self.counter

        sec = DummySecureShare()

        await _send_secure_shares_to_others(
            sec.message_body_generator,
            interpod_worker_mailbox,
        )

        # Call count is pods in `pod_mailbox_ids`
        # excluding this pod as it doesnt send to itself
        assert mock_message_service.send_message.call_count == 2
        mock_message_service.send_message.assert_has_calls(
            [
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.SECURE_SHARE,
                        body=msgpack.dumps(1),
                        recipient=other_pod_details[0].pod_identifier,
                        recipient_mailbox_id=other_pod_details[0].mailbox_id,
                        sender=pod_identifier,
                        sender_mailbox_id=worker_mailbox_id,
                        timestamp=fake_timestamps[0],
                        task_id=opt_task_id,
                    ),
                    already_packed=True,
                ),
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.SECURE_SHARE,
                        body=msgpack.dumps(2),
                        recipient=other_pod_details[1].pod_identifier,
                        recipient_mailbox_id=other_pod_details[1].mailbox_id,
                        sender=pod_identifier,
                        sender_mailbox_id=worker_mailbox_id,
                        timestamp=fake_timestamps[1],
                        task_id=opt_task_id,
                    ),
                    already_packed=True,
                ),
            ]
        )

    async def test__get_worker_secure_shares(
        self,
        caplog: LogCaptureFixture,
        interpod_worker_mailbox: _InterPodWorkerMailbox,
        mock_private_key: Mock,
        other_pod_details: List[WorkerDetails],
        patch_poll_for_messages: PollForMessagesPatcher,
    ) -> None:
        """Tests the _get_worker_secure_shares function works."""
        # Create mock messages to process
        mock_messages = [
            NonCallableMagicMock(
                spec=_BitfountMessage,
                **{
                    "sender": details.pod_identifier,
                    "message_type": _BitfountMessageType.SECURE_SHARE,
                    "decrypt_rsa.return_value.body": i,
                },
            )
            for i, details in enumerate(other_pod_details)
        ]
        patch_poll_for_messages(mock_messages)

        with caplog.at_level(logging.DEBUG):
            shares = await _run_func_and_listen_to_mailbox(
                _get_worker_secure_shares(interpod_worker_mailbox),
                interpod_worker_mailbox,
            )

        # Check decoded shares are what we expect and that each message was decoded
        assert shares == [i for i in range(len(other_pod_details))]
        for mock_message in mock_messages:
            mock_message.decrypt_rsa.assert_called_once_with(mock_private_key)

        # Check that we logged out the messages received
        debug_logs = get_debug_logs(caplog)
        for other_pod in other_pod_details:
            assert (
                f"Receiving secure share from worker {other_pod.pod_identifier}"
                in debug_logs
            )



"""Tests deployment config files can be loaded."""
import desert
from jinja2 import Environment, FileSystemLoader, select_autoescape
import pytest
import yaml

from bitfount.runners.config_schemas import PodConfig
from tests.integration import DEPLOYMENTS_DIRECTORY
from tests.utils.helper import integration_test


@integration_test
@pytest.mark.parametrize("config_file", ["census_income_config.yaml"])
def test_staging_deployment_configs_are_valid(config_file: str) -> None:
    """Test that the staging pod deployment configs are valid."""
    expected_username = "someTestUser"
    env = Environment(
        loader=FileSystemLoader(DEPLOYMENTS_DIRECTORY), autoescape=select_autoescape()
    )

    template = env.get_template(config_file + ".j2")
    rendered_config = template.render(
        env={"USERNAME": expected_username, "STAGING": "true"}
    )
    config_yaml = yaml.safe_load(rendered_config)
    config = desert.schema(PodConfig).load(config_yaml)

    assert config.username == expected_username
    assert "staging" in config.access_manager.url
    assert "staging" in config.hub.url
    assert "staging" in config.message_service.url


@integration_test
@pytest.mark.parametrize("config_file", ["census_income_config.yaml"])
def test_production_deployment_configs_are_valid(config_file: str) -> None:
    """Test that the production pod deployment configs are valid."""
    expected_username = "anotherTestUser"
    env = Environment(
        loader=FileSystemLoader(DEPLOYMENTS_DIRECTORY), autoescape=select_autoescape()
    )

    template = env.get_template(config_file + ".j2")
    rendered_config = template.render(env={"USERNAME": expected_username})
    config_yaml = yaml.safe_load(rendered_config)
    config = desert.schema(PodConfig).load(config_yaml)

    assert config.username == expected_username
    assert "staging" not in config.access_manager.url
    assert "staging" not in config.hub.url
    assert "staging" not in config.message_service.url



"""Vanilla model parameter aggregators for Federated Averaging."""
from __future__ import annotations

import logging
from typing import (
    Any,
    ClassVar,
    Iterable,
    List,
    Mapping,
    Optional,
    Union,
    cast,
    overload,
)

import numpy as np
import pandas as pd

from bitfount.federated.aggregators.base import (
    AggregationType,
    _AggregatorWorkerFactory,
    _BaseAggregatorFactory,
    _BaseModellerAggregator,
    _BaseWorkerAggregator,
)
from bitfount.federated.exceptions import AggregatorError
from bitfount.federated.shim import BackendTensorShim, _load_default_tensor_shim
from bitfount.types import (
    T_DTYPE,
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _SerializedWeights,
    _Weights,
)
from bitfount.utils import delegates

logger = logging.getLogger(__name__)


class _ModellerSide(_BaseModellerAggregator[T_DTYPE]):
    """Modeller-side of the vanilla aggregator."""

    def __init__(
        self,
        *,
        tensor_shim: BackendTensorShim,
        weights: Optional[Mapping[str, Union[float, int]]] = None,
        **kwargs: Any,
    ):
        """Create a new modeller-side instance of the vanilla aggregator.

        Args:
            tensor_shim: A shim providing methods to convert to/from
                tensor-like objects.
            weights: A mapping of pod identifiers to the desired weighting to give
                them in the averaging. If not supplied, all will be equally weighted.
                Weights will be normalised so that they sum to 1.0.
            **kwargs: Other keyword arguments.
        """
        super().__init__(tensor_shim=tensor_shim, **kwargs)
        self._weights: Optional[Mapping[str, float]] = self._normalise_weights(weights)

    @property
    def weights(self) -> Optional[Mapping[str, float]]:
        """The per-pod update weightings (for weighted average) for this aggregator.

        This attribute is a dictionary of pod identifiers to weights or `None` if no
        weights were specified.

        :::note

        This does not refer to the weights (parameters) of the model, but rather the
        weights to apply to the parameters of the model when performing a weighted
        average.

        :::
        """
        if self._weights:
            return dict(self._weights)
        else:
            return None

    @staticmethod
    def _normalise_weights(
        weights: Optional[Mapping[str, Union[float, int]]]
    ) -> Optional[Mapping[str, float]]:
        """Normalises the supplied weights to sum to 1.0.

        If no weights supplied, returns None.

        :::note

        This does not refer to the weights (parameters) of the model, but rather the
        weights to apply to the parameters of the model when performing a weighted
        average.

        :::
        """
        if not weights:
            return None

        weight_sum = sum(weights.values())
        return {pod_id: weight / weight_sum for pod_id, weight in weights.items()}

    def _validate_algorithm_outputs(
        self, algorithm_outputs: Iterable[Any], output_type: AggregationType
    ) -> None:
        """Validate the algorithm outputs.

        :::note

        This is used to validate that the algorithm outputs are of the rights type and
        are consistent between workers.

        :::
        """
        if output_type == AggregationType.UNSUPPORTED:
            raise AggregatorError(
                "Algorithm outputs are not recognised. Currently only pandas "
                "dataframes, numpy arrays and tensor state dictionaries are supported."
            )

        if output_type == AggregationType.TENSOR_DICT:
            algorithm_outputs = cast(Iterable[_SerializedWeights], algorithm_outputs)
            algo_outputs_iter = iter(algorithm_outputs)
            first = next(algo_outputs_iter)
            keys = first.keys()

            # Check that these are the same in all others
            if not all(keys == other.keys() for other in algo_outputs_iter):
                raise AggregatorError(
                    f"Keys are not consistent between workers: "
                    f"all names should match {sorted(set(keys))}"
                )

        elif output_type == AggregationType.PANDAS_DATAFRAME:
            # Check if all dataframes have the same columns and shapes
            # Also check if non-numeric columns are the same across all dataframes
            if len(set(i.shape for i in algorithm_outputs)) != 1:
                raise AggregatorError(
                    "Algorithm outputs are not consistent between workers: "
                    "all dataframes should have the same shape."
                )

            if len(set(tuple(i.columns) for i in algorithm_outputs)) != 1:
                raise AggregatorError(
                    "Algorithm outputs are not consistent between workers: "
                    "all dataframes should have the same columns."
                )

            self._set_non_numeric_columns(algorithm_outputs)

        elif output_type == AggregationType.NUMPY_ARRAY:

            if len(set(i.shape for i in algorithm_outputs)) != 1:
                raise AggregatorError(
                    "Algorithm outputs are not consistent between workers: "
                    "all arrays should have the same shape."
                )

            for output in algorithm_outputs:
                output = cast(np.ndarray, output)
                if not np.issubdtype(output.dtype, np.number):
                    raise AggregatorError(
                        f"Numpy array dtype {output.dtype} is not a number type."
                    )

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, _SerializedWeights],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> _Weights:
        """Averages tensor dictionaries and returns the average."""
        ...

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, pd.DataFrame],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Averages pandas dataframes and returns the average."""
        ...

    @overload
    def run(
        self,
        algorithm_outputs: Mapping[str, np.ndarray],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> np.ndarray:
        """Averages numpy arrays and returns the average."""
        ...

    def run(
        self,
        algorithm_outputs: Union[
            Mapping[str, _SerializedWeights],
            Mapping[str, np.ndarray],
            Mapping[str, pd.DataFrame],
        ],
        tensor_dtype: Optional[T_DTYPE] = None,
        **kwargs: Any,
    ) -> Union[_Weights, np.ndarray, pd.DataFrame]:
        """Averages algorithm outputs and returns them.

        Args:
            algorithm_outputs: A mapping of pod names to their algorithm outputs.
            tensor_dtype: The dtype to use when converting parameter updates to tensors.

        Raises:
            AggregatorError: If there is any issue whereby aggregation cannot proceed.
                Details will be included in the error message.
        """
        obj_type = self._get_type(algorithm_outputs.values())
        self._validate_algorithm_outputs(algorithm_outputs.values(), obj_type)
        # Use provided weights or, if none provided, use equal weights.
        weights = self._get_weights(algorithm_outputs)
        averaged_output: Union[pd.DataFrame, np.ndarray, _Weights]

        if obj_type == AggregationType.TENSOR_DICT:
            averaged_output = {}
            algorithm_outputs = cast(
                Mapping[str, _SerializedWeights], algorithm_outputs
            )
            state_dict_param_names = list(algorithm_outputs.values())[0].keys()
            for param_name in state_dict_param_names:
                averaged_output[param_name] = self._tensor_shim.to_tensor(
                    np.stack(
                        [
                            weights[pod_id] * np.asarray(params[param_name])
                            for pod_id, params in algorithm_outputs.items()
                        ],
                        axis=0,
                    ).sum(axis=0),
                    dtype=tensor_dtype,
                )
        elif obj_type == AggregationType.PANDAS_DATAFRAME:
            averaged_output = pd.DataFrame(self.non_numeric_columns)
            algorithm_outputs = cast(Mapping[str, pd.DataFrame], algorithm_outputs)
            df_columns: List[str] = list(algorithm_outputs.values())[0].columns

            for param_name in df_columns:
                if param_name not in self.non_numeric_columns:
                    averaged_output[param_name] = np.stack(
                        [
                            weights[pod_id] * params[param_name].to_numpy()
                            for pod_id, params in algorithm_outputs.items()
                        ],
                        axis=0,
                    ).sum(axis=0)

        elif obj_type == AggregationType.NUMPY_ARRAY:
            algorithm_outputs = cast(Mapping[str, np.ndarray], algorithm_outputs)
            averaged_output = np.stack(
                [weights[pod_id] * arr for pod_id, arr in algorithm_outputs.items()],
                axis=0,
            ).sum(axis=0)

        return averaged_output

    def _get_weights(self, algorithm_outputs: Mapping[str, Any]) -> Mapping[str, float]:
        """Gets the supplied weights or creates equal weights.

        Will raise appropriate errors if the pods in the supplied weights and
        the pods in the parameter updates don't match.
        """
        weights = self.weights

        if weights is None:
            # Use equal weights
            weights = {
                pod_id: 1 / len(algorithm_outputs)
                for pod_id in algorithm_outputs.keys()
            }
        else:
            # If using provided weights, check we have updates for each pod
            parameter_update_pods = set(algorithm_outputs.keys())
            weights_pods = set(weights.keys())

            extra_in_update = parameter_update_pods.difference(weights_pods)
            missing_in_update = weights_pods.difference(algorithm_outputs)

            if extra_in_update:
                raise AggregatorError(
                    f"Aggregation weightings provided but found updates from "
                    f"unweighted pods in received parameter updates: "
                    f"{';'.join(extra_in_update)}"
                )
            if missing_in_update:
                raise AggregatorError(
                    f"Aggregation weightings provided but missing updates from "
                    f"expected pods in received parameter updates: "
                    f"{';'.join(missing_in_update)}"
                )

        return weights


class _WorkerSide(_BaseWorkerAggregator):
    """Worker-side of the vanilla aggregator."""

    def __init__(self, *, tensor_shim: BackendTensorShim, **kwargs: Any):
        super().__init__(tensor_shim=tensor_shim, **kwargs)

    @overload
    async def run(
        self, algorithm_output: _Weights, **kwargs: Any
    ) -> _SerializedWeights:
        """Converts tensors to list of floats and returns them."""
        ...

    @overload
    async def run(self, algorithm_output: np.ndarray, **kwargs: Any) -> np.ndarray:
        """Doesn't do anything - leaves the type as it is."""
        ...

    @overload
    async def run(self, algorithm_output: pd.DataFrame, **kwargs: Any) -> pd.DataFrame:
        """Doesn't do anything - leaves the type as it is."""
        ...

    async def run(
        self, algorithm_output: Union[_Weights, np.ndarray, pd.DataFrame], **kwargs: Any
    ) -> Union[_SerializedWeights, np.ndarray, pd.DataFrame]:
        """Converts input to a serializable format if applicable and returns it.

        Args:
            algorithm_output: The algorithm output to be encoded and sent to the
                modeller.

        Raises:
            AggregatorError: If the algorithm output is not a supported type.
        """
        obj_type = self._get_type([algorithm_output])
        aggregator_output: Union[_SerializedWeights, pd.DataFrame, np.ndarray]

        if obj_type == AggregationType.TENSOR_DICT:
            algorithm_output = cast(_Weights, algorithm_output)
            aggregator_output = {
                name: self._tensor_shim.to_list(param)
                for name, param in algorithm_output.items()
            }
        elif obj_type in (
            AggregationType.PANDAS_DATAFRAME,
            AggregationType.NUMPY_ARRAY,
        ):
            # Pandas dataframes and numpy arrays are converted to lists of floats
            # at the transport layer so no need to deal with them here
            algorithm_output = cast(Union[pd.DataFrame, np.ndarray], algorithm_output)
            aggregator_output = algorithm_output
        else:
            raise AggregatorError("Unrecognised type received from algorithm.")

        return aggregator_output


@delegates()
class Aggregator(_BaseAggregatorFactory, _AggregatorWorkerFactory):
    """Vanilla model parameter aggregator for Federated Averaging.

    Performs simple arithmetic mean of unencrypted model parameters.

    :::danger

    This aggregator is not secure. Algorithm outputs are shared with participants in an
    unencrypted manner. It is not recommended to use this aggregator in a zero-trust
    setting.

    :::

    Attributes:
        name: The name of the aggregator.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._tensor_shim = _load_default_tensor_shim()

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the Aggregator."""
        return _ModellerSide(tensor_shim=self._tensor_shim, **kwargs)

    def worker(self, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the Aggregator."""
        return _WorkerSide(tensor_shim=self._tensor_shim, **kwargs)



"""Pytest configuration for model algorithm tests."""
from unittest.mock import Mock

from _pytest.fixtures import fixture


@fixture
def model() -> Mock:
    """Returns mock model."""
    mod = Mock(steps=10)
    mod.fit = Mock()
    mod.evaluate = Mock()
    mod.evaluate.return_value = ([0.9], [0.8])
    mod.backend_tensor_shim = Mock()
    return mod



"""Module for serialization and deserialization of models, algorithms and protocols."""
import itertools
from pathlib import Path
from typing import Any, Mapping, Optional, Tuple, Type, Union, cast

from marshmallow import Schema as MarshmallowSchema
from marshmallow import fields, post_load

from bitfount.federated.model_reference import BitfountModelReference
from bitfount.models.base_models import CNNModelStructure
from bitfount.schemas.exceptions import SchemaClassError
from bitfount.types import _JSONDict

# Methods for dumping the schema
# In bf_dump, we take an object which we convert to a schema using the
# `_obj_to_schema` function. For each item provided in to the
# `_obj_to_schema` function we also have to load all attributes
#  of the parent classes,which we do using `_bf_combine_dict_dump.


def bf_dump(bf_obj: Any) -> _JSONDict:
    """Method for serializing an object.

    Args:
        bf_obj: The object we want to generate the schema dump for.

    Raises:
        SchemaClassError: if the provided object is None.
    """
    # Get the schema class from the object.
    schema_cls = _obj_to_schema(bf_obj)

    if schema_cls is None:
        raise SchemaClassError(
            f"There is no schema class that can be "
            f"generated from the given object:{bf_obj}."
        )
    else:
        myschema = schema_cls()

        # Get the schema dump
        schema_dump = cast(_JSONDict, myschema.dump(bf_obj))
        # Make sure all fields and the schema are deleted to prevent
        # memory leak.
        del myschema

        return schema_dump


def _obj_to_schema(bf_obj: Any, depth: int = 0) -> Optional[Type[MarshmallowSchema]]:
    """Converts an object to a marshmallow schema."""
    if depth > 5:
        raise SchemaClassError("Cannot have more than depth 5 in serialisation")

    if bf_obj:
        # Load all nested and unnested attributes of the parent classes.
        unnested_fields, nested_fields = _bf_combine_dict_dump(
            bf_obj, bf_obj.fields_dict.copy(), bf_obj.nested_fields.copy()
        )
        unnested_fields["class_name"] = fields.Str()
        nested_fields_helper = {}

        # For each nested item, go recursively through them and unpack
        # to only have unnested fields and then generate schemas for
        # all of them.
        for k, _ in nested_fields.items():
            aux = {}
            # We only care about the case when `_obj_to_schema` is no None
            if _obj_to_schema(getattr(bf_obj, k), depth + 1) is not None:
                aux = {k: _obj_to_schema(getattr(bf_obj, k), depth + 1)}
            nested_fields_helper.update(aux)

        # Get the schema of the object
        return _get_marshmallow_schema(
            bf_obj.__class__, unnested_fields, nested_fields_helper
        )
    else:
        return None


def _bf_combine_dict_dump(
    cls: Any, fields_dict: _JSONDict, nested_fields: _JSONDict
) -> Tuple[_JSONDict, _JSONDict]:
    """Combine the nested and unnested fields from the subclasses.

    Loop through the class mro, and get the fields_dict and
    nested_fields from the parent classes.
    """
    for item in cls.__class__.__mro__:
        fields_dict, nested_fields = _combine_dict_helper(
            cls, item, fields_dict, nested_fields
        )
    return fields_dict, nested_fields


# Shared Methods


def _get_marshmallow_schema(
    obj_class: Any,
    unnested_fields: _JSONDict,
    nested_schemas: _JSONDict,
) -> Type[MarshmallowSchema]:
    """Generate the marshmallow schema from field dictionaries."""
    # For all nested items, define them as a nested field with the generated schema.
    nested = ((name, fields.Nested(value)) for (name, value) in nested_schemas.items())

    @post_load
    def recreate_factory(self: Any, data: _JSONDict, **_kwargs: Any) -> Any:
        data = data.copy()
        data.pop("class_name")
        return obj_class(**data)

    # Pack all fields needed for the schema generation
    all_fields = dict(
        itertools.chain.from_iterable(
            [
                nested,
                unnested_fields.items(),
                [(recreate_factory.__name__, recreate_factory)],
            ]
        )
    )
    # Get the return schema from the fields.
    ReturnSchema = _BitfountGeneratedSchema.from_dict(all_fields)
    return ReturnSchema


class _BitfountGeneratedSchema(MarshmallowSchema):
    """Schema class to prevent registering all the Marshmallow Schemas generated."""

    class Meta:
        """Meta class for the _BitfountGeneratedSchema.

        Used to set the register to `False`, to make sure that the newly generated
        schemas are not registered in the marshmallow class registers.
        """

        register = False

    def __del__(self) -> None:
        """Helper function for deleting all schema attributes.

        It makes sure that all the fields are assigned to None
        when schema gets deleted, to free up memory.
        """
        # Mypy complains that there is a type mismatch between these
        # fields and `None` type, however, our goal is to make sure
        # they are set to None when deleting the schema, so we can
        # ignore the assignment.
        self.fields = None  # type: ignore[assignment] # Reason: see above
        self.declared_fields = None  # type: ignore[assignment] # Reason: see above # noqa: B950
        self.load_fields = None  # type: ignore[assignment] # Reason: see above
        self.dump_fields = None  # type: ignore[assignment] # Reason: see above

    # The below methods are related to the `BitfountModelReference`.
    # They are used for serialization and deserialization of the model_ref.
    # The marshmallow field for the model_ref is using these methods for
    # serializing/ deserializing.
    @staticmethod
    def get_model_ref(bfmr: BitfountModelReference) -> str:
        """Returns the model_ref, ready for serialization.

        Used for serialization of BitfountModelReference.
        """
        model_ref = bfmr.model_ref
        try:
            return model_ref.stem  # type: ignore[union-attr]  # Reason: captured by AttributeError below  # noqa: B950
        except AttributeError as ae:
            # Check if class name only, return if is
            if Path(model_ref).stem == str(model_ref):
                return str(model_ref)
            # Otherwise error
            raise TypeError(
                f"Unable to serialise model_ref; "
                f"expected python file path Path or model name str, "
                f"got {type(model_ref)} with value {model_ref}"
            ) from ae

    @staticmethod
    def load_model_ref(value: str) -> Union[Path, str]:
        """Deserialize the model_ref value.

        Used for deserialization of BitfountModelReference.
        """
        try:
            new_value = Path(value).expanduser()
            if new_value.stem == str(new_value):  # i.e. is just a class name
                return str(value)
            return new_value
        except TypeError:
            return str(value)


def _combine_dict_helper(
    cls: Any, item: Any, fields_dict: _JSONDict, nested_fields: _JSONDict
) -> Tuple[_JSONDict, _JSONDict]:
    """Helper function for bf_combine_load and _dump.

    Used to get the nested and unnested fields from the `item` class.
    """
    if hasattr(item, "fields_dict"):
        for k, v in item.fields_dict.items():
            if k not in fields_dict:
                fields_dict[k] = v
        for k, v in item.nested_fields.items():
            if k not in nested_fields:
                nested_fields[k] = v
    if hasattr(item, "optional_nested") and item.optional_nested is not None:
        if hasattr(cls, "optional_nested") and cls.optional_nested is not None:
            aux = [v for v in item.optional_nested if v not in cls.optional_nested]
            cls.optional_nested.extend(aux)
        else:
            cls.optional_nested = item.optional_nested.copy()
    return fields_dict, nested_fields


# Methods for loading the schema

# In bf_load, we take a dictionary which we convert to a schema using the
# `_dict_to_schema` function. For each item provided in to the
# `_dict_to_schema` function we also have to loop through the registers
#  given to make sure we load all attributes of the parent classes,
#  which we do using `_bf_combine_dict_load`.


def bf_load(dct: _JSONDict, registry: Mapping[str, Any]) -> Any:
    """Method for deserializing an object.

    Args:
        dct: A JSON dictionary with the fields to load.
        registry: The registry where we can access the schema type.
    """
    # Get the schema given a dictionary with the fields.
    schema_cls = _dict_to_schema(dct, registry)
    # Get the schema class and load the schema
    myschema = schema_cls()
    schema_load = myschema.load(dct)
    # Make sure all fields and the schema are deleted to prevent
    # memory leak.
    del myschema

    return schema_load


def _dict_to_schema(
    dct: _JSONDict, registry: Mapping[str, type], depth: int = 0
) -> Any:
    """Converts a dictionary to a marshmallow schema."""
    if depth > 5:
        raise SchemaClassError("Cannot have more than depth 5 in serialisation")
    if dct:
        # Get the class by looking in the given registry for the class_name after
        # removing the `bitfount.` prefix if present. This is present only on protocols,
        # algorithms, aggregators and models.
        cls: Any = registry[dct["class_name"].split(".", 1)[-1]]
        # Load all nested and unnested attributes of the parent classes.
        unnested_fields, nested_fields = _bf_combine_dict_load(
            cls, cls.fields_dict.copy(), cls.nested_fields.copy(), registry
        )
        unnested_fields["class_name"] = fields.Str()
        nested_fields_aux = {}

        # For each nested item, go recursively through them and unpack
        # to only have unnested fields and then generate schemas for
        # all of them.
        for (k, sub_reg) in nested_fields.items():
            if hasattr(cls, "optional_nested") and cls.optional_nested is not None:
                # If the nested item is optional, we can skip the KeyError
                if k in cls.optional_nested:
                    try:
                        aux = {k: _dict_to_schema(dct[k], sub_reg, depth + 1)}
                    except KeyError:
                        continue
                else:
                    aux = {k: _dict_to_schema(dct[k], sub_reg, depth + 1)}
            else:
                cls.optional_nested = None
                aux = {k: _dict_to_schema(dct[k], sub_reg, depth + 1)}
            nested_fields_aux.update(aux)
        # Get the schema of the class
        return _get_marshmallow_schema(cls, unnested_fields, nested_fields_aux)
    else:
        return None


def _bf_combine_dict_load(
    cls: Any,
    fields_dict: _JSONDict,
    nested_fields: _JSONDict,
    registry: Mapping[str, Any],
) -> Tuple[_JSONDict, _JSONDict]:
    """Combine the nested and unnested fields from the subclasses.

    This method loops through the registry and checks if any of
    the modules are superclasses of the given class. This is
    done explicitly as some of the subclasses are not loaded
    properly in the federated setting. This way, it ensures
    that both nested and unnested fields are loaded from all
    the respective subclasses.

    Args:
        cls: the class for which we want the nested and unnested fields.
        fields_dict: the unnested fields dictionary.
        nested_fields: the nested fields dictionary.
        registry: the registry with the (possible) superclasses
    """
    # We need a special check
    # for subclasses of CNNModelStructure as it may
    # overrides the default layer value with the layer value defined in the
    # NeuralNetworkModelStructure.
    # Below we force it to take the default value of the CNNModelStructure,
    # if it has not been defined with the model.

    if issubclass(cls, CNNModelStructure):
        if "layers" not in cls.fields_dict:
            fields_dict["layers"] = [16, 32]

    # Loop through the registry in order to load all relevant
    # nested and unnested fields from the parent classes.
    for item in registry.values():
        if issubclass(cls, item):
            fields_dict, nested_fields = _combine_dict_helper(
                cls, item, fields_dict, nested_fields
            )
    return fields_dict, nested_fields



"""PyTorch implementations of the federated learning mixin classes."""
from __future__ import annotations

from abc import ABC
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, cast

import pytorch_lightning as pl
import torch

from bitfount.backends.pytorch.federated.shim import PyTorchBackendTensorShim
from bitfount.backends.pytorch.types import _AdaptorForPyTorchTensor
from bitfount.federated.mixins import _DistributedModelMixIn
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.shim import _load_default_tensor_shim
from bitfount.models.bitfount_model import BitfountModel
from bitfount.types import (
    T_DTYPE,
    DistributedModelProtocol,
    _Residuals,
    _SerializedWeights,
    _TensorLike,
    _Weights,
)

if TYPE_CHECKING:
    from bitfount.data.dataloaders import BitfountDataLoader
    from bitfount.hub.api import BitfountHub
    from bitfount.types import _DistributedModelTypeOrReference

logger = logging.getLogger(__name__)


class _PyTorchDistributedModelMixIn(_DistributedModelMixIn, ABC):
    """PyTorch implementation of the DistributedModelMixIn."""

    epochs: Optional[int]
    steps: Optional[int]
    train_dl: BitfountDataLoader
    _pl_trainer: pl.Trainer
    _total_num_batches_trained: int

    def get_param_states(self) -> Dict[str, _TensorLike]:
        """See base class.

        Wrapping the state dictionary with `dict` ensures we return a `dict` rather than
        an `OrderedDict`.
        """
        aux = dict(self._model.state_dict())  # type: ignore[attr-defined] # Reason: _model is initialised in subclass # noqa: B950

        return self._get_torch_adapter_states(aux)

    def _get_torch_tensor_states(
        self, adapted_params: Mapping[str, _TensorLike]
    ) -> Dict[str, torch.Tensor]:
        """Get the tensors out of our adapter."""
        tensor_dict: Dict[str, torch.Tensor] = {}
        for k, v in adapted_params.items():
            if isinstance(v, _AdaptorForPyTorchTensor):
                tensor_dict[k] = v.torchtensor
        return tensor_dict

    def _get_torch_adapter_states(
        self, torch_tensor_params: Mapping[str, torch.Tensor]
    ) -> Dict[str, _TensorLike]:
        """Put tensors in our torch.Tensor adapter."""
        return {k: _AdaptorForPyTorchTensor(v) for k, v in torch_tensor_params.items()}

    def apply_weight_updates(self, weight_updates: Sequence[_Weights]) -> _Weights:
        """See base class."""
        params_to_update_adapted = self.get_param_states()
        params_to_update = self._get_torch_tensor_states(params_to_update_adapted)
        tensor_weight_updates = [
            self._get_torch_tensor_states(params) for params in weight_updates
        ]
        weight = 1 / len(weight_updates)
        for name in tensor_weight_updates[0]:
            params_to_update[name].data.copy_(
                params_to_update[name]
                + torch.stack(
                    [weight * params[name].data for params in tensor_weight_updates],
                    dim=0,
                ).sum(dim=0)
            )
        adapted_params: _Weights = self._get_torch_adapter_states(params_to_update)
        return adapted_params

    def update_params(self, new_model_params: _Weights) -> None:
        """See base class."""
        current_params_adapted = self.get_param_states()
        current_params = self._get_torch_tensor_states(current_params_adapted)
        new_model_params_torch = self._get_torch_tensor_states(new_model_params)
        for name in new_model_params_torch:
            current_params[name].data.copy_(new_model_params_torch[name].data)

    def deserialize_params(self, serialized_weights: _SerializedWeights) -> _Weights:
        """Convert serialized model params to tensors."""
        backend_tensor_shim = cast(
            PyTorchBackendTensorShim, _load_default_tensor_shim()
        )

        weights = {
            name: backend_tensor_shim.to_tensor(param)
            for name, param in serialized_weights.items()
        }

        return weights

    def serialize_params(self, weights: _Weights) -> _SerializedWeights:
        """Serialize model params."""
        backend_tensor_shim = cast(
            PyTorchBackendTensorShim, _load_default_tensor_shim()
        )

        serialized_weights = {
            name: backend_tensor_shim.to_list(param) for name, param in weights.items()
        }

        return serialized_weights

    @staticmethod
    def diff_params(
        old_params: _Weights,
        new_params: _Weights,
    ) -> _Residuals:
        """See base class."""
        old_params_torch: Dict[str, torch.Tensor] = {}
        for k, v in old_params.items():
            if isinstance(v, _AdaptorForPyTorchTensor):
                old_params_torch[k] = v.torchtensor
        new_params_torch: Dict[str, torch.Tensor] = {}
        for k, v in new_params.items():
            if isinstance(v, _AdaptorForPyTorchTensor):
                new_params_torch[k] = v.torchtensor
        for name in new_params:
            old_params_torch[name].data.copy_(
                new_params_torch[name].data - old_params_torch[name].data
            )
        old_params_adapted: _Residuals = {
            k: _AdaptorForPyTorchTensor(v) for k, v in old_params_torch.items()
        }
        return old_params_adapted

    def set_model_training_iterations(self, iterations: int) -> None:
        """See base class."""
        if self.epochs:
            self.epochs = iterations
            if hasattr(self, "_pl_trainer"):
                self._pl_trainer.fit_loop.max_epochs = iterations
        else:
            self.steps = iterations
            if hasattr(self, "_pl_trainer"):
                self._pl_trainer.fit_loop.epoch_loop.max_steps = iterations

    def reset_trainer(self) -> None:
        """See base class."""
        # `trainer_init()` comes from `BasePyTorchModel` or `PyTorchBitfountModel`
        # This is a standard pattern for MixIn classes.
        self._pl_trainer: pl.Trainer = self.trainer_init()  # type: ignore[attr-defined] # Reason: see above. # noqa: B950

        # The `max_epochs` attribute on the trainer does not need to be reset for
        # epochs. This behaviour is only for steps. See pytorch-lightning issue #11425
        if self.steps:
            # `max_steps` must be set to however number of steps we want to train
            # greater than the number of batches already trained
            self._pl_trainer.fit_loop.epoch_loop.max_steps = self.steps + (
                self._total_num_batches_trained % len(self.train_dl)
            )

    def tensor_precision(self) -> T_DTYPE:
        """Returns torch default dtype.

        This is `torch.float32` by default unless changed. This method should be
        overridden in the subclass if the model supports non-32-bit model tensors.
        """
        return cast(T_DTYPE, torch.get_default_dtype())

    def log_(self, name: str, value: Any, **kwargs: Any) -> Any:
        """Simple wrapper around the pytorch lightning `log` method."""
        # Method is present on `BasePyTorchModel` inherited from `pl.LightningModule`.
        self.log(name, value, **kwargs)  # type: ignore[attr-defined] # Reason: see above. # noqa: B950

    def _get_model(
        self,
        extra_imports: Optional[List[str]] = None,
        hub: Optional[BitfountHub] = None,
        **kwargs: Any,
    ) -> _DistributedModelTypeOrReference:
        """Gets the model to be used in distributed training.

        This method is used by the Modeller to get the model to be used in
        distributed training.

        Args:
            extra_imports: A list of extra imports to add to the model file if the model
                is a custom model.
            hub: A BitfountHub instance to be passed to `BitfountModelReference`.
                Optional.
            **kwargs: Additional keyword arguments to pass to the
                `BitfountModelReference` constructor as hyperparameters.

        Returns:
            Either a `BitfountModelReference` object if the model is a custom model, or
            the model itself if it is a built-in model.
        """
        if isinstance(self, BitfountModel):
            # If the model is a custom model, we need to save the source code to a file
            # and create a BitfountModelReference to pass to the algorithm.
            filename = f"{type(self).__name__}.py"
            self.serialize_model_source_code(
                filename=filename, extra_imports=extra_imports
            )
            logger.info(f"Custom model source code saved to file: {filename}")
            return BitfountModelReference(
                model_ref=Path(filename),
                model_version=kwargs.get("model_version"),
                datastructure=kwargs.get("datastructure") or self.datastructure,
                schema=kwargs.get("schema") or self.schema,
                hub=hub,
                hyperparameters={"epochs": self.epochs, "steps": self.steps, **kwargs},
            )

        return cast(DistributedModelProtocol, self)



"""Tests datasource.py."""
import datetime
from functools import partial
from pathlib import Path
import re
from typing import Any, Callable, Dict
from unittest.mock import Mock, PropertyMock

from _pytest.tmpdir import TempPathFactory
from intermine.webservice import Service
import pandas as pd
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy
from sqlalchemy.exc import ArgumentError

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.csv_source import CSVSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datasources.excel_source import ExcelSource
from bitfount.data.datasources.intermine_source import IntermineSource
from bitfount.data.exceptions import DataNotLoadedError, ExcelSourceError
from bitfount.data.types import DataPathModifiers
from bitfount.data.utils import DatabaseConnection, _hash_str
from tests.utils import PytestRequest
from tests.utils.helper import (
    DATASET_ROW_COUNT,
    create_dataset,
    integration_test,
    unit_test,
)


@fixture(scope="module")
def dataframe() -> pd.DataFrame:
    """Dataframe fixture."""
    return create_dataset()


@unit_test
class TestBaseSource:
    """Tests core BaseSource functionality with a CSV file."""

    @fixture(scope="function", params=["pandas", "image"])
    def datasource_generator(
        self, request: PytestRequest
    ) -> Callable[..., DataFrameSource]:
        """Dataset loader for use in tests."""
        image = False
        if request.param == "image":
            image = True
        data = create_dataset(image=image)
        if image:
            return partial(DataFrameSource, data, seed=420)

        return partial(DataFrameSource, data, seed=420)

    def test_get_data_dtypes(self) -> None:
        """Tests _get_data_dtypes."""
        df = pd.DataFrame(
            {
                "int_column": [1, 2, 3],
                "float_column": [1.0, 2.0, 3.0],
                "str_column": ["1", "2", "3"],
            }
        )
        df["date"] = datetime.date(2020, 1, 1)
        df["datetime"] = datetime.datetime(2020, 1, 1, 0, 0)
        dtypes = BaseSource._get_data_dtypes(df)

        # Check that the correct columns are returned
        assert isinstance(dtypes, dict)
        assert sorted(list(dtypes)) == sorted(df.columns)
        # Check that the types are as expected
        assert dtypes["date"] == object
        assert dtypes["date"] != pd.StringDtype()
        assert dtypes["datetime"] != object
        assert dtypes["datetime"] == pd.StringDtype()

    def test_tabular_datasource_errors(self) -> None:
        """Checks BaseSource object errors via wrong first argument."""
        with pytest.raises(TypeError):
            DataFrameSource("test1", seed=420)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950

        with pytest.raises(TypeError):
            test_path = Path("/my/root/directory")
            DataFrameSource(test_path, seed=420)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950

    def test_datasource_modifiers_path_prefix(self, dataframe: pd.DataFrame) -> None:
        """Tests functionality for providing image path prefix."""
        dataframe["image"] = "image_file_name"
        modifiers = {"image": DataPathModifiers({"prefix": "/path/to/"})}
        datasource = DataFrameSource(dataframe, seed=420, modifiers=modifiers)
        datasource.load_data()
        assert len(datasource.data["image"].unique()) == 1
        assert datasource.data["image"].unique()[0] == "/path/to/image_file_name"

    def test_image_datasource_ext_suffix(self, dataframe: pd.DataFrame) -> None:
        """Tests functionality for finding images by file extension."""
        dataframe["image"] = "image_file_name"
        modifiers = {"image": DataPathModifiers({"suffix": ".jpeg"})}
        datasource = DataFrameSource(dataframe, seed=420, modifiers=modifiers)
        datasource.load_data()
        assert len(datasource.data["image"].unique()) == 1
        assert datasource.data["image"].unique()[0] == "image_file_name.jpeg"

    def test_image_datasource_ext_prefix_suffix(self, dataframe: pd.DataFrame) -> None:
        """Tests functionality for finding images by file extension."""
        dataframe["image"] = "image_file_name"
        modifiers = {
            "image": DataPathModifiers({"prefix": "/path/to/", "suffix": ".jpeg"})
        }
        datasource = DataFrameSource(dataframe, seed=420, modifiers=modifiers)
        datasource.load_data()
        assert len(datasource.data["image"].unique()) == 1
        assert datasource.data["image"].unique()[0] == "/path/to/image_file_name.jpeg"

    def test_multiple_img_datasource_modifiers(self) -> None:
        """Tests functionality for finding multiple images by file extension."""
        data = create_dataset(multiimage=True, img_size=1)
        data["image1"] = "image1_file_name"
        data["image2"] = "image2_file_name"
        modifiers = {
            "image1": DataPathModifiers({"prefix": "/path/to/"}),
            "image2": DataPathModifiers({"suffix": ".jpeg"}),
        }
        datasource = DataFrameSource(data, seed=420, modifiers=modifiers)
        datasource.load_data()
        assert len(datasource.data["image1"].unique()) == 1
        assert datasource.data["image1"].unique()[0] == "/path/to/image1_file_name"
        assert len(datasource.data["image2"].unique()) == 1
        assert datasource.data["image2"].unique()[0] == "image2_file_name.jpeg"

    def test_tabular_datasource_read_csv_correctly(
        self, dataframe: pd.DataFrame, tmp_path: Path
    ) -> None:
        """Tests CSVSource loading from csv."""
        file_path = tmp_path / "tabular_data_test.csv"
        dataframe.to_csv(file_path)
        ds = CSVSource(file_path)
        ds.load_data()
        assert ds.data is not None

    def test_ignored_cols_list_excluded_from_df(self, dataframe: pd.DataFrame) -> None:
        """Tests that a list of ignore_cols are ignored in the data."""
        dataframe["image"] = "image_file_name"
        ignore_cols = ["N", "O", "P"]
        datasource = DataFrameSource(
            dataframe,
            seed=420,
            ignore_cols=ignore_cols,
        )
        datasource.load_data()
        assert not any(item in datasource.data.columns for item in ignore_cols)

    def test_ignored_single_col_list_excluded_from_df(
        self, dataframe: pd.DataFrame
    ) -> None:
        """Tests that a str ignore_cols is ignored in the data."""
        dataframe["image"] = "image_file_name"
        ignore_cols = "N"
        datasource = DataFrameSource(
            dataframe,
            seed=420,
            ignore_cols=ignore_cols,
        )
        datasource.load_data()
        assert ignore_cols not in datasource.data.columns

    def test_hash(
        self,
        datasource_generator: Callable[..., DataFrameSource],
        mocker: MockerFixture,
    ) -> None:
        """Tests hash is called on the dtypes."""
        datasource = datasource_generator()
        expected_hash = f"hash_{id(datasource._table_hashes)}"
        mock_hash_function: Mock = mocker.patch(
            "bitfount.data.datasources.base_source._generate_dtypes_hash",
            return_value=expected_hash,
            autospec=True,
        )
        datasource.get_dtypes()

        actual_hash = datasource.hash

        # Check hash is expected return and how it was called
        assert actual_hash == _hash_str(str([expected_hash]))
        mock_hash_function.assert_called_once()

    def test_get_dtypes_ignores_cols(
        self,
        datasource_generator: Callable[..., DataFrameSource],
        mocker: MockerFixture,
    ) -> None:
        """Tests get_dtypes drops _ignore_cols."""
        datasource = datasource_generator()
        datasource._ignore_cols = ["A"]
        assert "A" in datasource.get_data().columns

        result = datasource.get_dtypes()

        assert "A" not in result.keys()

    def test_get_column_applies_modifiers(
        self,
        datasource_generator: Callable[..., DataFrameSource],
        mocker: MockerFixture,
    ) -> None:
        """Tests get_column applies modifiers."""
        datasource = datasource_generator()
        prefix = "/path/to/"
        datasource._modifiers = {"A": DataPathModifiers({"prefix": prefix})}
        expected_result = prefix + datasource.get_data()["A"].astype(str)

        result = datasource.get_column("A")

        assert all(result == expected_result)

    def test_get_data_failes(self, mock_engine: Mock) -> None:
        """Test data raises error when data not set."""
        db_conn = DatabaseConnection(mock_engine)
        datasource = DatabaseSource(db_conn)
        with pytest.raises(
            DataNotLoadedError,
            match="Data is not loaded yet. Please call `load_data` first.",
        ):
            datasource.data


class TestDatabaseSource:
    """Tests DatabaseSource."""

    @unit_test
    @pytest.mark.parametrize(
        "params, result",
        [
            ({"table_names": ["dummy_data"]}, "SELECT * FROM dummy_data"),
            ({"query": "SELECT * FROM dummy_data"}, "SELECT * FROM dummy_data"),
        ],
    )
    def test_query(
        self, mock_engine: Mock, params: Dict[Any, Any], result: Any
    ) -> None:
        """Test query returns correct result."""
        db_conn = DatabaseConnection(mock_engine, **params)
        datasource = DatabaseSource(db_conn, seed=420)
        assert datasource.query == result

    @unit_test
    def test_query_datastructure_query(self, mock_engine: Mock) -> None:
        """Test query returns datastructure query."""
        db_conn = DatabaseConnection(
            mock_engine,
        )
        datastructure_query = "SELECT * FROM dummy_data"
        datasource = DatabaseSource(db_conn, seed=420)
        datasource.datastructure_query = datastructure_query
        assert datasource.query == datastructure_query

    @unit_test
    def test_hash_multitable_raises_value_error(self, mock_engine: Mock) -> None:
        """Tests hash function raises `DataNotLoadedError` if data is not loaded."""
        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        datasource = DatabaseSource(db_conn, seed=420)
        with pytest.raises(DataNotLoadedError):
            datasource.hash

    @unit_test
    def test_value_error_raised_if_no_table_name_provided_for_multitable_datasource(
        self, mock_engine: Mock
    ) -> None:
        """Test ValueError raised if no table_name for multi-table DatabaseSource."""
        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        ds.load_data()
        with pytest.raises(
            ValueError, match="No table name provided for multi-table datasource."
        ):
            ds.get_dtypes()

    @unit_test
    def test_value_error_raised_if_table_not_found_for_multitable_datasource(
        self, mock_engine: Mock
    ) -> None:
        """Tests ValueError raised if table missing for multi-table DatabaseSource."""
        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        ds.load_data()
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Table name not_a_table not found in the data. "
                + "Available tables: ['dummy_data', 'dummy_data_2']"
            ),
        ):
            ds.get_dtypes(table_name="not_a_table")

    @unit_test
    def test_invalid_connection_string_errors(self) -> None:
        """Test DatabaseSource errors with invalid conn string."""
        invalid_connection_str = "random_str"
        with pytest.raises(
            ArgumentError,
            match=(
                f"Invalid db_conn. db_conn: {invalid_connection_str} "
                "must be sqlalchemy compatible database url, see: .*"
            ),
        ):
            DatabaseSource(invalid_connection_str)

    @integration_test
    def test_instantiation_works_with_db_conn_as_string(
        self, db_session: sqlalchemy.engine.base.Engine
    ) -> None:
        """Test DataSource can be created with valid connection string."""
        expected_table_names = ["dummy_data", "dummy_data_2"]
        connection_str = str(db_session.url)
        assert isinstance(connection_str, str)
        ds = DatabaseSource(connection_str)

        # Correctly read table from connection string
        assert ds.table_names == expected_table_names

    @integration_test
    def test_mock_get_dtypes_reads_and_returns_table_schema(
        self, db_session: sqlalchemy.engine.base.Engine
    ) -> None:
        """Tests that the `get_dtypes` method returns a dictionary.

        Also checks that the dtypes hash is added appropriately.
        """
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)

        assert len(ds._table_hashes) == 0
        assert isinstance(ds.get_dtypes(table_name="dummy_data"), dict)
        assert len(ds._table_hashes) == 1

    @integration_test
    def test_get_dtypes_reads_and_returns_table_schema(
        self, db_session: sqlalchemy.engine.base.Engine
    ) -> None:
        """Tests that the `get_dtypes` method returns a dictionary.

        Also checks that the dtypes hash is added appropriately.
        """
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        assert len(ds._table_hashes) == 0
        table = ds.get_dtypes(table_name="dummy_data")
        assert isinstance(table, dict)
        assert len(ds._table_hashes) == 1

    @unit_test
    def test_get_column(self, mock_engine: Mock, mocker: MockerFixture) -> None:
        """Test get_column returns column."""
        # Creates a multitable DatabaseConnection object
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        datasource = DatabaseSource(db_conn)
        col_name = "A"
        mock_table: Mock = mocker.patch(
            "bitfount.data.datasources.database_source.Table",
            autospec=True,
        )
        mock_table.return_value.columns = {col_name: None}
        mock_session: Mock = mocker.patch(
            "bitfount.data.datasources.database_source.Session",
            autospec=True,
        )
        mock_result = [(1,), (2,), (3,)]
        mock_session.return_value.__enter__.return_value.query.return_value = (
            mock_result
        )

        result = datasource.get_column(col_name=col_name, table_name="dummy_data")

        assert all(result == pd.Series([1, 2, 3]))
        mock_table.assert_called_once()
        mock_session.assert_called_once()

    @unit_test
    def test_len_magic_method(self, mock_engine: Mock, mocker: MockerFixture) -> None:
        """Tests that __len__ magic method returns correct row count."""
        # Mocks `execute` method on the SQLAlchemy connection object and the
        # `scalar_one` method on the resulting cursor result to return the
        # dataset row count
        mock_db_connection = Mock()
        mock_result = Mock()
        mock_result.scalar_one.return_value = DATASET_ROW_COUNT
        mock_db_connection.execute.return_value = mock_result
        mock_engine.execution_options.return_value = mock_engine

        # Creates a multitable DatabaseConnection object
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        # Mocks `connect` method and resulting context manager on SQLAlchemy Engine
        mocker.patch.object(
            db_conn.con, "connect"
        ).return_value.__enter__.return_value = mock_db_connection
        loader = DatabaseSource(db_conn)

        # Calls __len__ method on loader
        dataset_length = len(loader)

        # Makes assertions on call stack in order
        # Ignoring mypy errors because `connect` has been patched to return a Mock
        db_conn.con.connect.assert_called_once()  # type: ignore[attr-defined] # Reason: see above # noqa: B950
        db_conn.con.connect.return_value.__enter__.assert_called_once()  # type: ignore[attr-defined]  # Reason: see above # noqa: B950
        mock_db_connection.execute.assert_called_once()
        mock_result.scalar_one.assert_called_once()

        # Makes assertion on final result
        assert dataset_length == DATASET_ROW_COUNT

    @unit_test
    def test_get_dtypes_raises_value_error_if_table_name_is_none(
        self, mock_engine: Mock
    ) -> None:
        """Tests that ValueError is raised if there is no table name provided."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        datasource = DatabaseSource(db_conn)
        with pytest.raises(
            ValueError, match="No table name provided for multi-table datasource."
        ):
            datasource.get_dtypes(table_name=None)

    @unit_test
    def test_get_column_raises_value_error_if_table_name_is_none(
        self, mock_engine: Mock
    ) -> None:
        """Tests that ValueError is raised if there is no table name provided."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        datasource = DatabaseSource(db_conn)
        with pytest.raises(
            ValueError, match="No table name provided for multi-table datasource."
        ):
            datasource.get_column(col_name="col", table_name=None)

    @unit_test
    def test_get_values_raises_value_error_if_table_name_is_none(
        self, mock_engine: Mock
    ) -> None:
        """Tests that ValueError is raised if there is no table name provided."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        datasource = DatabaseSource(db_conn)
        with pytest.raises(
            ValueError, match="No table name provided for multi-table datasource."
        ):
            datasource.get_values(table_name=None, col_names=["col1", "col2"])

    @unit_test
    def test_validate_table_name_raises_value_error_if_tables_dont_exist(
        self, mock_engine: Mock
    ) -> None:
        """Tests that ValueError is raised if there are no tables."""
        db_conn = DatabaseConnection(
            mock_engine,
            query="DUMMY QUERY",
        )
        loader = DatabaseSource(db_conn)
        with pytest.raises(
            ValueError, match="Database Connection is not aware of any tables."
        ):
            loader._validate_table_name("dummy_data")

    @unit_test
    def test_validate_table_name_raises_value_error_if_table_name_not_found(
        self, mock_engine: Mock
    ) -> None:
        """Tests that ValueError is raised if the table name is not found."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        loader = DatabaseSource(db_conn)
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Table name blah not found in the data. "
                "Available tables: ['dummy_data', 'dummy_data_2']",
            ),
        ):
            loader._validate_table_name("blah")


@unit_test
class TestCSVSource:
    """Tests CSVSource."""

    @fixture
    def csv_source(self, dataframe: pd.DataFrame, tmp_path: Path) -> CSVSource:
        """CSVSource."""
        file_path = tmp_path / "tabular_data_test.csv"
        dataframe.to_csv(file_path, index=False)
        datasource = CSVSource(file_path, read_csv_kwargs={"parse_dates": ["Date"]})
        return datasource

    def test_len(self, csv_source: CSVSource) -> None:
        """Tests that __len__ magic method returns correct row count."""
        assert len(csv_source) == DATASET_ROW_COUNT

    def test_get_data(self, csv_source: CSVSource, dataframe: pd.DataFrame) -> None:
        """Test get_data returns dataframe."""
        dataframe["Date"] = pd.to_datetime(dataframe["Date"])
        result = csv_source.get_data()
        pd.testing.assert_frame_equal(dataframe, result, check_dtype=False)

    def test_get_column(self, csv_source: CSVSource, dataframe: pd.DataFrame) -> None:
        """Test get_column returns column."""
        column = "A"
        result = csv_source.get_column(column)
        assert all(dataframe[column] == result)

    def test_get_dtypes(self, csv_source: CSVSource, dataframe: pd.DataFrame) -> None:
        """Test get_dtypes works."""
        result = csv_source.get_dtypes()
        assert isinstance(result, dict)
        for col in dataframe.columns:
            assert col in result.keys()

    def test_multitable(self, csv_source: CSVSource) -> None:
        """Test multi_table for CSVSource."""
        assert not csv_source.multi_table


@unit_test
class TestDataFrameSource:
    """Tests DataFrameSource."""

    def test_len(self, dataframe: pd.DataFrame) -> None:
        """Tests that __len__ magic method returns correct row count."""
        loader = DataFrameSource(dataframe)
        assert len(loader) == DATASET_ROW_COUNT


@unit_test
class TestIntermineSource:
    """Tests IntermineSource."""

    @fixture
    def intermine_source(
        self, dataframe: pd.DataFrame, mocker: MockerFixture
    ) -> IntermineSource:
        """IntermineSource."""
        service_url = "https://beta.humanmine.org/humanmine"
        mock_service = mocker.patch(
            "bitfount.data.datasources.intermine_source.Service"
        )
        mock_service.get_template_by_user.return_value = Service(
            "https://beta.humanmine.org/humanmine"
        ).get_template("Gene_Protein")
        datasource = IntermineSource(service_url, None)
        mocker.patch.object(
            IntermineSource, "table_names", PropertyMock(return_value=["table_name"])
        )
        mocker.patch.object(IntermineSource, "_template_to_df", return_value=dataframe)

        datasource.template_to_user_map = {"table_name": "user1"}
        return datasource

    def test_len(self, intermine_source: IntermineSource) -> None:
        """Tests that __len__ magic method returns correct row count."""
        intermine_source.load_data(table_name="table_name")
        assert len(intermine_source) == DATASET_ROW_COUNT

    def test_get_data(self, intermine_source: IntermineSource) -> None:
        """Test get_data returns dataframe."""
        result = intermine_source.get_data(table_name="table_name")
        assert isinstance(result, pd.DataFrame)

    def test__validate_table_name_no_table_name(
        self, intermine_source: IntermineSource
    ) -> None:
        """Test get_data raises ValueError when no table_name specified."""
        with pytest.raises(
            ValueError, match="No table name provided for Intermine service."
        ):
            intermine_source._validate_table_name()

    def test_get_data_service_with_no_templates(
        self, intermine_source: IntermineSource, mocker: MockerFixture
    ) -> None:
        """Test get_data raises ValueError the service has no templates."""
        mock_table_names = PropertyMock(return_value=[])
        mocker.patch.object(IntermineSource, "table_names", mock_table_names)
        with pytest.raises(
            ValueError, match="Service .* did not return any templates."
        ):
            intermine_source.get_data(table_name="table_name")

    def test_get_data_service_with_invalid_table_name(
        self, intermine_source: IntermineSource
    ) -> None:
        """Test get_data raises ValueError with invalid table name."""
        with pytest.raises(
            ValueError,
            match=(
                "Template name invalid_table_name not found in service: .*. "
                "Available tables: .*"
            ),
        ):
            intermine_source.get_data(table_name="invalid_table_name")

    def test_get_column(self, intermine_source: IntermineSource) -> None:
        """Test get_column returns column."""
        column = "A"
        result = intermine_source.get_column(column, table_name="table_name")
        assert isinstance(result, pd.Series)

    def test_get_column_no_table_name(self, intermine_source: IntermineSource) -> None:
        """Test get_column raises ValueError when no table_name specified."""
        column = "Gene.symbol"
        with pytest.raises(ValueError, match="Expected parameter: table_name."):
            intermine_source.get_column(column)

    def test_get_dtypes(self, intermine_source: IntermineSource) -> None:
        """Test get_dtypes works."""
        result = intermine_source.get_dtypes(table_name="table_name")
        assert isinstance(result, dict)

    def test_get_values(self, intermine_source: IntermineSource) -> None:
        """Test get_values works."""
        result = intermine_source.get_values(col_names=["A"], table_name="table_name")
        assert isinstance(result, dict)

    def test_multitable(
        self, intermine_source: IntermineSource, mocker: MockerFixture
    ) -> None:
        """Test multi_table for IntermineSource."""
        mock_table_names = PropertyMock(return_value=["table_1", "table_2"])
        mocker.patch.object(IntermineSource, "table_names", mock_table_names)
        assert intermine_source.multi_table

    def test_multitable_false_for_single_template(
        self, intermine_source: IntermineSource
    ) -> None:
        """Test multi_table for single table IntermineSource."""
        assert not intermine_source.multi_table

    def test__check_duplicate_templates(self, mocker: MockerFixture) -> None:
        """Test ValueError raised when there are duplicate template names."""
        service_url = "https://beta.humanmine.org/humanmine"
        mock_service = mocker.patch(
            "bitfount.data.datasources.intermine_source.Service"
        )
        duplicate_table_names = ["table_name"] * 2
        mocker.patch.object(
            IntermineSource,
            "table_names",
            PropertyMock(return_value=duplicate_table_names),
        )
        expected_err = (
            "Duplicated template name: 'table_name', found in service. "
            "Template names must have unique names."
        )
        with pytest.raises(ValueError, match=expected_err):
            IntermineSource(service_url, None)
        mock_service.assert_called_once_with(service_url, token=None)


@unit_test
class TestExcelSource:
    """Tests ExcelSource."""

    @fixture(scope="class")
    def single_table_excel_file(
        self, tmp_path_factory: TempPathFactory, dataframe: pd.DataFrame
    ) -> Path:
        """Path to single table excel file."""
        tmp_path = tmp_path_factory.mktemp("temp_excel")
        filename = tmp_path / "test.xlsx"
        dataframe.to_excel(filename, index=False, sheet_name="Sheet1")
        return filename

    @fixture(scope="class")
    def multi_table_excel_file(
        self, tmp_path_factory: TempPathFactory, dataframe: pd.DataFrame
    ) -> Path:
        """Path to multi table excel file."""
        tmp_path = tmp_path_factory.mktemp("temp_excel")
        filename = tmp_path / "test.xlsx"
        with pd.ExcelWriter(filename) as writer:  # type: ignore[abstract] # Reason: This is the documented usage. # noqa: B950
            dataframe.to_excel(writer, index=False, sheet_name="Sheet1")
            dataframe.to_excel(writer, index=False, sheet_name="Sheet2")

        return filename

    @fixture(scope="class")
    def single_table_excel_source(self, single_table_excel_file: Path) -> ExcelSource:
        """Single Table ExcelSource."""
        source = ExcelSource(single_table_excel_file, sheet_name="Sheet1")
        assert not source.multi_table
        return source

    @fixture(scope="class")
    def multi_table_excel_source(self, multi_table_excel_file: Path) -> ExcelSource:
        """Multi Table ExcelSource."""
        source = ExcelSource(multi_table_excel_file, sheet_name=["Sheet1", "Sheet2"])
        assert source.multi_table
        return source

    def test_excel_source_raises_type_error_if_wrong_file_extension(self) -> None:
        """Test ExcelSource raises TypeError if wrong file extension."""
        with pytest.raises(
            TypeError, match="Please provide a Path or URL to an Excel file."
        ):
            ExcelSource("test.txt")

    def test_multi_table_excel_source_raises_value_error_if_column_names_provided(
        self, multi_table_excel_file: Path
    ) -> None:
        """Test Multi Table ExcelSource raises ValueError if column names provided."""
        with pytest.raises(
            ValueError,
            match="Column names can only be provided if a single sheet name is provided.",  # noqa: B950
        ):
            ExcelSource(
                multi_table_excel_file,
                sheet_name=["Sheet1", "Sheet2"],
                column_names=["A"],
            )

    @pytest.mark.parametrize("multi_table", [True, False])
    def test_excel_source_raises_value_error_if_referenced_sheets_are_missing(
        self,
        multi_table: bool,
        multi_table_excel_file: Path,
        single_table_excel_file: Path,
    ) -> None:
        """Test ExcelSource raises ValueError if referenced sheets are missing."""
        with pytest.raises(
            ValueError,
            match=re.escape("Sheet(s) Sheet3 were not found in the Excel file."),
        ):
            if multi_table:
                ExcelSource(
                    multi_table_excel_file,
                    sheet_name=["Sheet1", "Sheet2", "Sheet3"],  # Sheet3 is missing
                )
            else:
                ExcelSource(
                    single_table_excel_file,
                    sheet_name=["Sheet3"],  # Sheet3 is missing
                )

    def test_column_names_override_the_ones_in_the_excel_file(
        self, single_table_excel_file: Path
    ) -> None:
        """Test column names override the ones in the excel file."""
        new_column_names = [str(i) for i in range(16)]
        datasource = ExcelSource(
            single_table_excel_file,
            column_names=new_column_names,
            read_excel_kwargs={"skiprows": 1},
        )

        df = datasource.get_data()
        assert df is not None
        assert list(df.columns) == new_column_names

    def test_multi_table_get_data_raises_value_error_if_table_name_not_recognised(
        self, multi_table_excel_source: ExcelSource
    ) -> None:
        """Test Multi-Table ExcelSource raises ValueError if table name not recognised."""  # noqa: B950
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Table name Table3 not found in the data. "
                "Available tables: Sheet1, Sheet2"
            ),
        ):
            multi_table_excel_source.get_data(table_name="Table3")

    @pytest.mark.parametrize("multi_table", [True, False])
    def test_get_values(
        self,
        multi_table: bool,
        multi_table_excel_source: ExcelSource,
        single_table_excel_source: ExcelSource,
    ) -> None:
        """Test get_values method works as expected."""
        if multi_table:
            values = multi_table_excel_source.get_values(["A"], "Sheet1")
        else:
            values = single_table_excel_source.get_values(["A"])

        assert isinstance(values, dict)
        assert len(values["A"]) == len(set(values["A"]))  # type: ignore[arg-type] # Reason: Len methods is available. # noqa: B950

    @pytest.mark.parametrize("multi_table", [True, False])
    def test_get_column(
        self,
        multi_table: bool,
        multi_table_excel_source: ExcelSource,
        single_table_excel_source: ExcelSource,
    ) -> None:
        """Test get_column method works as expected."""
        if multi_table:
            values = multi_table_excel_source.get_column("A", "Sheet1")
        else:
            values = single_table_excel_source.get_column("A")

        assert isinstance(values, pd.Series)
        assert len(values) == DATASET_ROW_COUNT

    @pytest.mark.parametrize("multi_table", [True, False])
    def test_get_dtypes(
        self,
        multi_table: bool,
        multi_table_excel_source: ExcelSource,
        single_table_excel_source: ExcelSource,
    ) -> None:
        """Test get_dtypes method works as expected."""
        if multi_table:
            dtypes = multi_table_excel_source.get_dtypes("Sheet1")
        else:
            dtypes = single_table_excel_source.get_dtypes()

        assert isinstance(dtypes, dict)

    def test_multitable_get_dtypes_error_no_table(
        self,
        multi_table_excel_source: ExcelSource,
    ) -> None:
        """Test that error is raised when no table is provided."""
        with pytest.raises(ExcelSourceError):
            multi_table_excel_source.get_dtypes()

    @pytest.mark.parametrize("multi_table", [True, False])
    def test_len_magic_method(
        self,
        multi_table: bool,
        multi_table_excel_source: ExcelSource,
        single_table_excel_source: ExcelSource,
    ) -> None:
        """Test len magic method works as expected."""
        if multi_table:
            with pytest.raises(
                ValueError, match="Can't ascertain length of multi-table Excel dataset."
            ):
                len(multi_table_excel_source)

            multi_table_excel_source.load_data(table_name="Sheet1")
            length = len(multi_table_excel_source)
        else:
            length = len(single_table_excel_source)

        assert length == DATASET_ROW_COUNT

    @unit_test
    def test_get_values_raises_value_error_if_table_name_is_none(
        self, multi_table_excel_source: ExcelSource
    ) -> None:
        """Tests that ValueError is raised if there is no table name provided."""
        with pytest.raises(
            ValueError, match="No table name provided for multi-table datasource."
        ):
            multi_table_excel_source.get_values(
                table_name=None, col_names=["col1", "col2"]
            )

    @unit_test
    def test_get_column_raises_value_error_if_table_name_is_none(
        self, multi_table_excel_source: ExcelSource
    ) -> None:
        """Tests that ValueError is raised if there is no table name provided."""
        with pytest.raises(
            ValueError, match="No table name provided for multi-table datasource."
        ):
            multi_table_excel_source.get_column(table_name=None, col_name="col1")
